# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from warnings import warn
from numba.core import types, config
from numba.core.errors import DeprecationError, NumbaInvalidConfigWarning
from numba.cuda.compiler import declare_device_function
from numba.cuda.core import sigutils
from numba.cuda.dispatcher import CUDADispatcher
from numba.cuda.simulator.kernel import FakeCUDAKernel
from numba.cuda.cudadrv.driver import _have_nvjitlink


_msg_deprecated_signature_arg = (
    "Deprecated keyword argument `{0}`. "
    "Signatures should be passed as the first "
    "positional argument."
)


def jit(
    func_or_sig=None,
    device=False,
    inline="never",
    forceinline=False,
    link=[],
    debug=None,
    opt=None,
    lineinfo=False,
    cache=False,
    launch_bounds=None,
    lto=None,
    **kws,
):
    """
    JIT compile a Python function for CUDA GPUs.

    :param func_or_sig: A function to JIT compile, or *signatures* of a
       function to compile. If a function is supplied, then a
       :class:`Dispatcher <numba.cuda.dispatcher.CUDADispatcher>` is returned.
       Otherwise, ``func_or_sig`` may be a signature or a list of signatures,
       and a function is returned. The returned function accepts another
       function, which it will compile and then return a :class:`Dispatcher
       <numba.cuda.dispatcher.CUDADispatcher>`. See :ref:`jit-decorator` for
       more information about passing signatures.

       .. note:: A kernel cannot have any return value.
    :param device: Indicates whether this is a device function.
    :type device: bool
    :param inline: Enables inlining at the Numba IR level when set to
       ``"always"``. See `Notes on Inlining
       <https://numba.readthedocs.io/en/stable/developer/inlining.html>`_.
    :type inline: str
    :param forceinline: Enables inlining at the NVVM IR level when set to
       ``True``. This is accomplished by adding the ``alwaysinline`` function
       attribute to the function definition.
    :type forceinline: bool
    :param link: A list of files containing PTX or CUDA C/C++ source to link
       with the function
    :type link: list
    :param debug: If True, check for exceptions thrown when executing the
       kernel. Since this degrades performance, this should only be used for
       debugging purposes. If set to True, then ``opt`` should be set to False.
       Defaults to False.  (The default value can be overridden by setting
       environment variable ``NUMBA_CUDA_DEBUGINFO=1``.)
    :param fastmath: When True, enables fastmath optimizations as outlined in
       the :ref:`CUDA Fast Math documentation <cuda-fast-math>`.
    :param max_registers: Request that the kernel is limited to using at most
       this number of registers per thread. The limit may not be respected if
       the ABI requires a greater number of registers than that requested.
       Useful for increasing occupancy.
    :param opt: Whether to compile with optimization enabled. If unspecified,
       the OPT configuration variable is decided by ``NUMBA_OPT```; all
       non-zero values will enable optimization.
    :type opt: bool
    :param lineinfo: If True, generate a line mapping between source code and
       assembly code. This enables inspection of the source code in NVIDIA
       profiling tools and correlation with program counter sampling.
    :type lineinfo: bool
    :param cache: If True, enables the file-based cache for this function.
    :type cache: bool
    :param launch_bounds: Kernel launch bounds, specified as a scalar or a tuple
                          of between one and three items. Tuple items provide:

                          - The maximum number of threads per block,
                          - The minimum number of blocks per SM,
                          - The maximum number of blocks per cluster.

                          If a scalar is provided, it is used as the maximum
                          number of threads per block.
    :type launch_bounds: int | tuple[int]
    :param lto: Whether to enable LTO. If unspecified, LTO is enabled by
                default when nvjitlink is available, except for kernels where
                ``debug=True``.
    :type lto: bool
    """

    if link and config.ENABLE_CUDASIM:
        raise NotImplementedError("Cannot link PTX in the simulator")

    if kws.get("boundscheck"):
        raise NotImplementedError("bounds checking is not supported for CUDA")

    if kws.get("argtypes") is not None:
        msg = _msg_deprecated_signature_arg.format("argtypes")
        raise DeprecationError(msg)
    if kws.get("restype") is not None:
        msg = _msg_deprecated_signature_arg.format("restype")
        raise DeprecationError(msg)
    if kws.get("bind") is not None:
        msg = _msg_deprecated_signature_arg.format("bind")
        raise DeprecationError(msg)

    if isinstance(inline, bool):
        DeprecationWarning(
            "Passing bool to inline argument is deprecated, please refer to "
            "Numba's documentation on inlining: "
            "https://numba.readthedocs.io/en/stable/developer/inlining.html. "
            "You may have wanted the forceinline argument instead, to force "
            "inlining at the NVVM IR level."
        )

        inline = "always" if inline else "never"

    debug = config.CUDA_DEBUGINFO_DEFAULT if debug is None else debug
    opt = (config.OPT != 0) if opt is None else opt
    fastmath = kws.get("fastmath", False)
    extensions = kws.get("extensions", [])

    if debug and opt:
        msg = (
            "debug=True with opt=True "
            "is not supported by CUDA. This may result in a crash"
            " - set debug=False or opt=False."
        )
        warn(NumbaInvalidConfigWarning(msg))

    if debug and lineinfo:
        msg = (
            "debug and lineinfo are mutually exclusive. Use debug to get "
            "full debug info (this disables some optimizations), or "
            "lineinfo for line info only with code generation unaffected."
        )
        warn(NumbaInvalidConfigWarning(msg))

    if device and kws.get("link"):
        raise ValueError("link keyword invalid for device function")

    if lto is None:
        # Default to using LTO if nvjitlink is available and we're not debugging
        lto = _have_nvjitlink() and not debug
    else:
        if lto and not _have_nvjitlink():
            raise RuntimeError(
                "LTO requires nvjitlink, which is not available"
                "or not sufficiently recent (>=12.3)"
            )

    if sigutils.is_signature(func_or_sig):
        signatures = [func_or_sig]
        specialized = True
    elif isinstance(func_or_sig, list):
        signatures = func_or_sig
        specialized = False
    else:
        signatures = None

    if signatures is not None:
        if config.ENABLE_CUDASIM:

            def jitwrapper(func):
                return FakeCUDAKernel(func, device=device, fastmath=fastmath)

            return jitwrapper

        def _jit(func):
            targetoptions = kws.copy()
            targetoptions["debug"] = debug
            targetoptions["lineinfo"] = lineinfo
            targetoptions["link"] = link
            targetoptions["opt"] = opt
            targetoptions["fastmath"] = fastmath
            targetoptions["device"] = device
            targetoptions["inline"] = inline
            targetoptions["forceinline"] = forceinline
            targetoptions["extensions"] = extensions
            targetoptions["launch_bounds"] = launch_bounds
            targetoptions["lto"] = lto

            disp = CUDADispatcher(func, targetoptions=targetoptions)

            if cache:
                disp.enable_caching()

            for sig in signatures:
                argtypes, restype = sigutils.normalize_signature(sig)

                if restype and not device and restype != types.void:
                    raise TypeError("CUDA kernel must have void return type.")

                if device:
                    from numba.core import typeinfer

                    with typeinfer.register_dispatcher(disp):
                        disp.compile_device(argtypes, restype)
                else:
                    disp.compile(argtypes)

            disp._specialized = specialized
            disp.disable_compile()

            return disp

        return _jit
    else:
        if func_or_sig is None:
            if config.ENABLE_CUDASIM:

                def autojitwrapper(func):
                    return FakeCUDAKernel(
                        func, device=device, fastmath=fastmath
                    )
            else:

                def autojitwrapper(func):
                    return jit(
                        func,
                        device=device,
                        inline=inline,
                        forceinline=forceinline,
                        debug=debug,
                        opt=opt,
                        lineinfo=lineinfo,
                        link=link,
                        cache=cache,
                        launch_bounds=launch_bounds,
                        **kws,
                    )

            return autojitwrapper
        # func_or_sig is a function
        else:
            if config.ENABLE_CUDASIM:
                return FakeCUDAKernel(
                    func_or_sig, device=device, fastmath=fastmath
                )
            else:
                targetoptions = kws.copy()
                targetoptions["debug"] = debug
                targetoptions["lineinfo"] = lineinfo
                targetoptions["opt"] = opt
                targetoptions["link"] = link
                targetoptions["fastmath"] = fastmath
                targetoptions["device"] = device
                targetoptions["inline"] = inline
                targetoptions["forceinline"] = forceinline
                targetoptions["extensions"] = extensions
                targetoptions["launch_bounds"] = launch_bounds
                targetoptions["lto"] = lto
                disp = CUDADispatcher(func_or_sig, targetoptions=targetoptions)

                if cache:
                    disp.enable_caching()

                return disp


def declare_device(name, sig, link=None, use_cooperative=False):
    """
    Declare the signature of a foreign function. Returns a descriptor that can
    be used to call the function from a Python kernel.

    :param name: The name of the foreign function.
    :type name: str
    :param sig: The Numba signature of the function.
    :param link: External code to link when calling the function.
    :param use_cooperative: External code requires cooperative launch.
    """
    if link is None:
        link = tuple()
    else:
        if not isinstance(link, (list, tuple, set)):
            link = (link,)

    argtypes, restype = sigutils.normalize_signature(sig)
    if restype is None:
        msg = "Return type must be provided for device declarations"
        raise TypeError(msg)

    template = declare_device_function(
        name, restype, argtypes, link, use_cooperative
    )

    return template.key



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from numba import config
from . import enums

if config.CUDA_USE_NVIDIA_BINDING:
    from cuda.bindings.driver import CUjitInputType

    FILE_EXTENSION_MAP = {
        "o": CUjitInputType.CU_JIT_INPUT_OBJECT,
        "ptx": CUjitInputType.CU_JIT_INPUT_PTX,
        "a": CUjitInputType.CU_JIT_INPUT_LIBRARY,
        "lib": CUjitInputType.CU_JIT_INPUT_LIBRARY,
        "cubin": CUjitInputType.CU_JIT_INPUT_CUBIN,
        "fatbin": CUjitInputType.CU_JIT_INPUT_FATBINARY,
        "ltoir": CUjitInputType.CU_JIT_INPUT_NVVM,
    }
else:
    FILE_EXTENSION_MAP = {
        "o": enums.CU_JIT_INPUT_OBJECT,
        "ptx": enums.CU_JIT_INPUT_PTX,
        "a": enums.CU_JIT_INPUT_LIBRARY,
        "lib": enums.CU_JIT_INPUT_LIBRARY,
        "cubin": enums.CU_JIT_INPUT_CUBIN,
        "fatbin": enums.CU_JIT_INPUT_FATBINARY,
        "ltoir": enums.CU_JIT_INPUT_NVVM,
    }



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""
Declarations of the Runtime API functions.
"""

from ctypes import c_int, POINTER

API_PROTOTYPES = {
    # cudaError_t cudaRuntimeGetVersion ( int* runtimeVersion )
    "cudaRuntimeGetVersion": (c_int, POINTER(c_int)),
}



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from numba import cuda


@cuda.jit(device=True)
def cuda_module_in_device_function():
    return cuda.threadIdx.x



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from numba import cuda, float32
from numba.cuda.bf16 import bfloat16
from numba.cuda.testing import CUDATestCase

import math


class TestBfloat16HighLevelBindings(CUDATestCase):
    def skip_unsupported(self):
        if not cuda.is_bfloat16_supported():
            self.skipTest("bfloat16 requires compute capability 8.0+")

    def test_use_type_in_kernel(self):
        self.skip_unsupported()

        @cuda.jit
        def kernel():
            bfloat16(3.14)

        kernel[1, 1]()

    def test_math_bindings(self):
        self.skip_unsupported()

        exp_functions = [math.exp]
        try:
            from math import exp2

            exp_functions += [exp2]
        except ImportError:
            pass

        functions = [
            math.trunc,
            math.ceil,
            math.floor,
            math.sqrt,
            math.log,
            math.log10,
            math.cos,
            math.sin,
            math.tanh,
        ] + exp_functions

        for f in functions:
            with self.subTest(func=f):

                @cuda.jit
                def kernel(arr):
                    x = bfloat16(3.14)
                    y = f(x)
                    arr[0] = float32(y)

                arr = cuda.device_array((1,), dtype="float32")
                kernel[1, 1](arr)

                if f in exp_functions:
                    self.assertAlmostEqual(arr[0], f(3.14), delta=1e-1)
                else:
                    self.assertAlmostEqual(arr[0], f(3.14), delta=1e-2)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""
Added for symmetry with the core API
"""

from numba.core.extending import intrinsic as _intrinsic
from numba.cuda.models import register_model  # noqa: F401
from numba.cuda import models  # noqa: F401

intrinsic = _intrinsic(target="cuda")


def make_attribute_wrapper(typeclass, struct_attr, python_attr):
    """
    Make an automatic attribute wrapper exposing member named *struct_attr*
    as a read-only attribute named *python_attr*.
    The given *typeclass*'s model must be a StructModel subclass.

    Vendored from numba.core.extending with a change to consider the CUDA data
    model manager.
    """
    from numba.cuda.typing.templates import AttributeTemplate

    from numba.core.datamodel import default_manager
    from numba.core.datamodel.models import StructModel
    from numba.core.imputils import impl_ret_borrowed
    from numba.core import types
    from numba.cuda import cgutils

    from numba.cuda.models import cuda_data_manager
    from numba.cuda.cudadecl import registry as cuda_registry
    from numba.cuda.cudaimpl import registry as cuda_impl_registry

    data_model_manager = cuda_data_manager.chain(default_manager)

    if not isinstance(typeclass, type) or not issubclass(typeclass, types.Type):
        raise TypeError(f"typeclass should be a Type subclass, got {typeclass}")

    def get_attr_fe_type(typ):
        """
        Get the Numba type of member *struct_attr* in *typ*.
        """
        model = data_model_manager.lookup(typ)
        if not isinstance(model, StructModel):
            raise TypeError(
                f"make_attribute_wrapper() needs a type with a StructModel, but got {model}"
            )
        return model.get_member_fe_type(struct_attr)

    @cuda_registry.register_attr
    class StructAttribute(AttributeTemplate):
        key = typeclass

        def generic_resolve(self, typ, attr):
            if attr == python_attr:
                return get_attr_fe_type(typ)

    @cuda_impl_registry.lower_getattr(typeclass, python_attr)
    def struct_getattr_impl(context, builder, typ, val):
        val = cgutils.create_struct_proxy(typ)(context, builder, value=val)
        attrty = get_attr_fe_type(typ)
        attrval = getattr(val, struct_attr)
        return impl_ret_borrowed(context, builder, attrty, attrval)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import numpy as np
import ctypes
from numba.cuda.cudadrv.devicearray import (
    DeviceRecord,
    from_record_like,
    auto_device,
)
from numba.cuda.testing import unittest, CUDATestCase
from numba.cuda.testing import skip_on_cudasim
from numba.np import numpy_support
from numba import cuda

N_CHARS = 5

recordtype = np.dtype(
    [
        ("a", np.float64),
        ("b", np.int32),
        ("c", np.complex64),
        ("d", (np.str_, N_CHARS)),
    ],
    align=True,
)

recordwitharray = np.dtype([("g", np.int32), ("h", np.float32, 2)], align=True)

recwithmat = np.dtype([("i", np.int32), ("j", np.float32, (3, 3))])

recwithrecwithmat = np.dtype([("x", np.int32), ("y", recwithmat)])


@skip_on_cudasim("Device Record API unsupported in the simulator")
class TestCudaDeviceRecord(CUDATestCase):
    """
    Tests the DeviceRecord class with np.void host types.
    """

    def setUp(self):
        super().setUp()
        self._create_data(np.zeros)

    def _create_data(self, array_ctor):
        self.dtype = np.dtype([("a", np.int32), ("b", np.float32)], align=True)
        self.hostz = array_ctor(1, self.dtype)[0]
        self.hostnz = array_ctor(1, self.dtype)[0]
        self.hostnz["a"] = 10
        self.hostnz["b"] = 11.0

    def _check_device_record(self, reference, rec):
        self.assertEqual(rec.shape, tuple())
        self.assertEqual(rec.strides, tuple())
        self.assertEqual(rec.dtype, reference.dtype)
        self.assertEqual(rec.alloc_size, reference.dtype.itemsize)
        self.assertIsNotNone(rec.gpu_data)
        self.assertNotEqual(rec.device_ctypes_pointer, ctypes.c_void_p(0))

        numba_type = numpy_support.from_dtype(reference.dtype)
        self.assertEqual(rec._numba_type_, numba_type)

    def test_device_record_interface(self):
        hostrec = self.hostz.copy()
        devrec = DeviceRecord(self.dtype)
        self._check_device_record(hostrec, devrec)

    def test_device_record_copy(self):
        hostrec = self.hostz.copy()
        devrec = DeviceRecord(self.dtype)
        devrec.copy_to_device(hostrec)

        # Copy back and check values are all zeros
        hostrec2 = self.hostnz.copy()
        devrec.copy_to_host(hostrec2)
        np.testing.assert_equal(self.hostz, hostrec2)

        # Copy non-zero values to GPU and back and check values
        hostrec3 = self.hostnz.copy()
        devrec.copy_to_device(hostrec3)

        hostrec4 = self.hostz.copy()
        devrec.copy_to_host(hostrec4)
        np.testing.assert_equal(hostrec4, self.hostnz)

    def test_from_record_like(self):
        # Create record from host record
        hostrec = self.hostz.copy()
        devrec = from_record_like(hostrec)
        self._check_device_record(hostrec, devrec)

        # Create record from device record and check for distinct data
        devrec2 = from_record_like(devrec)
        self._check_device_record(devrec, devrec2)
        self.assertNotEqual(devrec.gpu_data, devrec2.gpu_data)

    def test_auto_device(self):
        # Create record from host record
        hostrec = self.hostnz.copy()
        devrec, new_gpu_obj = auto_device(hostrec)
        self._check_device_record(hostrec, devrec)
        self.assertTrue(new_gpu_obj)

        # Copy data back and check it is equal to auto_device arg
        hostrec2 = self.hostz.copy()
        devrec.copy_to_host(hostrec2)
        np.testing.assert_equal(hostrec2, hostrec)


class TestCudaDeviceRecordWithRecord(TestCudaDeviceRecord):
    """
    Tests the DeviceRecord class with np.record host types
    """

    def setUp(self):
        CUDATestCase.setUp(self)
        self._create_data(np.recarray)


@skip_on_cudasim("Structured array attr access not supported in simulator")
class TestRecordDtypeWithStructArrays(CUDATestCase):
    """
    Test operation of device arrays on structured arrays.
    """

    def _createSampleArrays(self):
        self.sample1d = cuda.device_array(3, dtype=recordtype)
        self.samplerec1darr = cuda.device_array(1, dtype=recordwitharray)[0]
        self.samplerecmat = cuda.device_array(1, dtype=recwithmat)[0]

    def setUp(self):
        super().setUp()
        self._createSampleArrays()

        ary = self.sample1d
        for i in range(ary.size):
            x = i + 1
            ary[i]["a"] = x / 2
            ary[i]["b"] = x
            ary[i]["c"] = x * 1j
            ary[i]["d"] = str(x) * N_CHARS

    def test_structured_array1(self):
        ary = self.sample1d
        for i in range(self.sample1d.size):
            x = i + 1
            self.assertEqual(ary[i]["a"], x / 2)
            self.assertEqual(ary[i]["b"], x)
            self.assertEqual(ary[i]["c"], x * 1j)
            self.assertEqual(ary[i]["d"], str(x) * N_CHARS)

    def test_structured_array2(self):
        ary = self.samplerec1darr
        ary["g"] = 2
        ary["h"][0] = 3.0
        ary["h"][1] = 4.0
        self.assertEqual(ary["g"], 2)
        self.assertEqual(ary["h"][0], 3.0)
        self.assertEqual(ary["h"][1], 4.0)

    def test_structured_array3(self):
        ary = self.samplerecmat
        mat = np.array(
            [[5.0, 10.0, 15.0], [20.0, 25.0, 30.0], [35.0, 40.0, 45.0]],
            dtype=np.float32,
        ).reshape(3, 3)
        ary["j"][:] = mat
        np.testing.assert_equal(ary["j"], mat)

    def test_structured_array4(self):
        arr = np.zeros(1, dtype=recwithrecwithmat)
        d_arr = cuda.to_device(arr)
        d_arr[0]["y"]["i"] = 1
        self.assertEqual(d_arr[0]["y"]["i"], 1)
        d_arr[0]["y"]["j"][0, 0] = 2.0
        self.assertEqual(d_arr[0]["y"]["j"][0, 0], 2.0)


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

# Contents in this file are referenced from the sphinx-generated docs.
# "magictoken" is used for markers as beginning and ending of example text.

import unittest
from numba.cuda.testing import CUDATestCase, skip_on_cudasim
from numba.cuda.tests.support import skip_unless_cffi, override_config


@skip_unless_cffi
@skip_on_cudasim("cudasim doesn't support cuda import at non-top-level")
class TestFFI(CUDATestCase):
    def test_ex_linking_cu(self):
        # magictoken.ex_linking_cu.begin
        from numba import cuda
        import numpy as np
        import os

        # Path to the source containing the foreign function
        # (here assumed to be in a subdirectory called "ffi")
        basedir = os.path.dirname(os.path.abspath(__file__))
        functions_cu = os.path.join(basedir, "ffi", "functions.cu")

        # Declaration of the foreign function
        mul = cuda.declare_device(
            "mul_f32_f32", "float32(float32, float32)", link=functions_cu
        )

        # A kernel that calls mul; functions.cu is linked automatically due to
        # the call to mul.
        @cuda.jit
        def multiply_vectors(r, x, y):
            i = cuda.grid(1)

            if i < len(r):
                r[i] = mul(x[i], y[i])

        # Generate random data
        N = 32
        np.random.seed(1)
        x = np.random.rand(N).astype(np.float32)
        y = np.random.rand(N).astype(np.float32)
        r = np.zeros_like(x)

        # Run the kernel
        multiply_vectors[1, 32](r, x, y)

        # Sanity check - ensure the results match those expected
        np.testing.assert_array_equal(r, x * y)
        # magictoken.ex_linking_cu.end

    def test_ex_from_buffer(self):
        from numba import cuda
        import os

        basedir = os.path.dirname(os.path.abspath(__file__))
        functions_cu = os.path.join(basedir, "ffi", "functions.cu")

        # magictoken.ex_from_buffer_decl.begin
        signature = "float32(CPointer(float32), int32)"
        sum_reduce = cuda.declare_device(
            "sum_reduce", signature, link=functions_cu
        )
        # magictoken.ex_from_buffer_decl.end

        # magictoken.ex_from_buffer_kernel.begin
        import cffi

        ffi = cffi.FFI()

        @cuda.jit
        def reduction_caller(result, array):
            array_ptr = ffi.from_buffer(array)
            result[()] = sum_reduce(array_ptr, len(array))

        # magictoken.ex_from_buffer_kernel.end

        import numpy as np

        x = np.arange(10).astype(np.float32)
        r = np.ndarray((), dtype=np.float32)

        reduction_caller[1, 1](r, x)

        expected = np.sum(x)
        actual = r[()]
        np.testing.assert_allclose(expected, actual)

    def test_ex_extra_includes(self):
        import numpy as np
        from numba import cuda, config
        import os

        basedir = os.path.dirname(os.path.abspath(__file__))
        mul_dir = os.path.join(basedir, "ffi", "include")
        saxpy_cu = os.path.join(basedir, "ffi", "saxpy.cu")

        testdir = os.path.dirname(basedir)
        add_dir = os.path.join(testdir, "data", "include")

        includedir = ":".join([mul_dir, add_dir])
        with override_config("CUDA_NVRTC_EXTRA_SEARCH_PATHS", includedir):
            # magictoken.ex_extra_search_paths.begin
            from numba import config

            includedir = ":".join([mul_dir, add_dir])
            config.CUDA_NVRTC_EXTRA_SEARCH_PATHS = includedir
            # magictoken.ex_extra_search_paths.end

            # magictoken.ex_extra_search_paths_kernel.begin
            sig = "float32(float32, float32, float32)"
            saxpy = cuda.declare_device("saxpy", sig=sig, link=saxpy_cu)

            @cuda.jit
            def vector_saxpy(a, x, y, res):
                i = cuda.grid(1)
                if i < len(res):
                    res[i] = saxpy(a, x[i], y[i])

            # magictoken.ex_extra_search_paths_kernel.end

            size = 10_000
            a = 3.0
            X = np.ones((size,), dtype="float32")
            Y = np.ones((size,), dtype="float32")
            R = np.zeros((size,), dtype="float32")

            block_size = 32
            num_blocks = (size // block_size) + 1

            vector_saxpy[num_blocks, block_size](a, X, Y, R)

            expected = a * X + Y
            np.testing.assert_equal(R, expected)


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import numpy as np

from numba import cuda
import unittest
from numba.cuda.testing import CUDATestCase


@cuda.jit
def foo(x):
    i = cuda.grid(1)
    if i < x.size:
        x[i] += 1


class TestForAll(CUDATestCase):
    def test_forall_1(self):
        arr = np.arange(11)
        orig = arr.copy()
        foo.forall(arr.size)(arr)
        np.testing.assert_array_almost_equal(arr, orig + 1)

    def test_forall_2(self):
        @cuda.jit("void(float32, float32[:], float32[:])")
        def bar(a, x, y):
            i = cuda.grid(1)
            if i < x.size:
                y[i] = a * x[i] + y[i]

        x = np.arange(13, dtype=np.float32)
        y = np.arange(13, dtype=np.float32)
        oldy = y.copy()
        a = 1.234
        bar.forall(y.size)(a, x, y)
        np.testing.assert_array_almost_equal(y, a * x + oldy, decimal=3)

    def test_forall_no_work(self):
        # Ensure that forall doesn't launch a kernel with no blocks when called
        # with 0 elements. See Issue #5017.
        arr = np.arange(11)
        foo.forall(0)(arr)

    def test_forall_negative_work(self):
        # Ensure that forall doesn't allow the creation of a forall with a
        # negative element count.
        with self.assertRaises(ValueError) as raises:
            foo.forall(-1)
        self.assertIn(
            "Can't create ForAll with negative task count",
            str(raises.exception),
        )


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import numpy as np
from numba import cuda
from numba.cuda.testing import unittest, CUDATestCase


class TestCudaAutoContext(CUDATestCase):
    def test_auto_context(self):
        """A problem was revealed by a customer that the use cuda.to_device
        does not create a CUDA context.
        This tests the problem
        """
        A = np.arange(10, dtype=np.float32)
        newA = np.empty_like(A)
        dA = cuda.to_device(A)

        dA.copy_to_host(newA)
        self.assertTrue(np.allclose(A, newA))


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import numba.cuda as cuda
from numba.cuda.testing import CUDATestCase, skip_on_cudasim
import llvmlite
from numba import types

"""
llvmlite pre 45 left redundant metadata nodes for debug info
for basic scalar types. Jie Li fixed this in llvmlite 45:
https://github.com/numba/llvmlite/pull/1165

This makes it difficult for us to use the same check patterns for both
llvmlite versions, as basic scalar types like int8 are emitted only once
per module. Some checks share most of the logic, but some are entirely
split between the two versions. The checks indicate which version of llvmlite
they are for, if they cannot be shared.
"""

scalar_types: tuple[tuple[types.Type, str]] = (
    (
        types.float32,
        """
        CHECK-LLVMLITE-LE44: distinct !DISubprogram
        CHECK: [[DBG86:.+]] = !DIBasicType(encoding: DW_ATE_float, name: "float32", size: 32)
        CHECK: !DILocalVariable(
        CHECK-SAME: name: "a"
        CHECK-SAME: type: [[DBG86]]
        """,
    ),
    (
        types.float64,
        """
        CHECK-LLVMLITE-LE44: distinct !DISubprogram
        CHECK: [[DBG86:.+]] = !DIBasicType(encoding: DW_ATE_float, name: "float64", size: 64)
        CHECK: !DILocalVariable(
        CHECK-SAME: name: "a"
        CHECK-SAME: type: [[DBG86]]
        """,
    ),
    (
        types.int8,
        """
        CHECK-LLVMLITE-LE44: distinct !DISubprogram
        CHECK: [[DBG86:.+]] = !DIBasicType(encoding: DW_ATE_signed, name: "int8", size: 8)
        CHECK: !DILocalVariable(
        CHECK-SAME: name: "a"
        CHECK-SAME: type: [[DBG86]]
        """,
    ),
    (
        types.int16,
        """
        CHECK-LLVMLITE-LE44: distinct !DISubprogram
        CHECK: [[DBG86:.+]] = !DIBasicType(encoding: DW_ATE_signed, name: "int16", size: 16)
        CHECK: !DILocalVariable(
        CHECK-SAME: name: "a"
        CHECK-SAME: type: [[DBG86]]
        """,
    ),
    (
        types.int32,
        """
        CHECK-LLVMLITE-LE44: distinct !DISubprogram
        CHECK: [[DBG86:.+]] = !DIBasicType(encoding: DW_ATE_signed, name: "int32", size: 32)
        CHECK: !DILocalVariable(
        CHECK-SAME: name: "a"
        CHECK-SAME: type: [[DBG86]]
        """,
    ),
    (
        types.int64,
        """
        CHECK-LLVMLITE-LE44: distinct !DISubprogram
        CHECK: [[DBG86:.+]] = !DIBasicType(encoding: DW_ATE_signed, name: "int64", size: 64)
        CHECK: !DILocalVariable(
        CHECK-SAME: name: "a"
        CHECK-SAME: type: [[DBG86]]
        """,
    ),
    (
        types.uint8,
        """
        CHECK-LLVMLITE-LE44: distinct !DISubprogram
        CHECK: [[DBG86:.+]] = !DIBasicType(encoding: DW_ATE_unsigned
        CHECK-SAME: name: "uint8", size: 8)
        CHECK: !DILocalVariable(
        CHECK-SAME: name: "a"
        CHECK-SAME: type: [[DBG86]]
        """,
    ),
    (
        types.uint16,
        """
        CHECK-LLVMLITE-LE44: distinct !DISubprogram
        CHECK: [[DBG86:.+]] = !DIBasicType(encoding: DW_ATE_unsigned, name: "uint16", size: 16)
        CHECK: !DILocalVariable(
        CHECK-SAME: name: "a"
        CHECK-SAME: type: [[DBG86]]
        """,
    ),
    (
        types.uint32,
        """
        CHECK-LLVMLITE-LE44: distinct !DISubprogram
        CHECK: [[DBG86:.+]] = !DIBasicType(encoding: DW_ATE_unsigned, name: "uint32", size: 32)
        CHECK: !DILocalVariable(
        CHECK-SAME: name: "a"
        CHECK-SAME: type: [[DBG86]]
        """,
    ),
    (
        types.uint64,
        """
        CHECK-LLVMLITE-LE44: distinct !DISubprogram
        CHECK: [[DBG86:.+]] = !DIBasicType(encoding: DW_ATE_unsigned, name: "uint64", size: 64)
        CHECK: !DILocalVariable(
        CHECK-SAME: name: "a"
        CHECK-SAME: type: [[DBG86]]
        """,
    ),
    (
        types.complex64,
        """
        CHECK-LLVMLITE-LE44: distinct !DISubprogram
        CHECK-LLVMLITE-LE44: [[DBG98:.+]] = !DIBasicType(encoding: DW_ATE_float, name: "float", size: 32)
        CHECK-LLVMLITE-LE44: [[DBG99:.+]] = !DIDerivedType(
        CHECK-LLVMLITE-LE44-SAME: baseType: [[DBG98]]
        CHECK-LLVMLITE-LE44-SAME: name: "real"
        CHECK-LLVMLITE-LE44-SAME: offset: 0
        CHECK-LLVMLITE-LE44-SAME: size: 32
        CHECK-LLVMLITE-LE44-SAME: tag: DW_TAG_member
        CHECK-LLVMLITE-LE44-SAME: )
        CHECK-LLVMLITE-LE44: [[DBG100:.+]] = !DIBasicType(encoding: DW_ATE_float, name: "float", size: 32)
        CHECK-LLVMLITE-LE44: [[DBG101:.+]] = !DIDerivedType(
        CHECK-LLVMLITE-LE44-SAME: baseType: [[DBG100]]
        CHECK-LLVMLITE-LE44-SAME: name: "imag"
        CHECK-LLVMLITE-LE44-SAME: offset: 32
        CHECK-LLVMLITE-LE44-SAME: size: 32
        CHECK-LLVMLITE-LE44-SAME: tag: DW_TAG_member
        CHECK-LLVMLITE-LE44-SAME: )
        CHECK-LLVMLITE-LE44: [[DBG102:.+]] = !{ [[DBG99]], [[DBG101]] }
        CHECK-LLVMLITE-LE44: [[DBG103:.+]] = distinct !DICompositeType(
        CHECK-LLVMLITE-LE44-SAME: elements: [[DBG102]]
        CHECK-LLVMLITE-LE44-SAME: identifier: "{float, float}"
        CHECK-LLVMLITE-LE44-SAME: name: "complex64 ({float, float})"
        CHECK-LLVMLITE-LE44-SAME: size: 64
        CHECK-LLVMLITE-LE44-SAME: tag: DW_TAG_structure_type
        CHECK-LLVMLITE-LE44-SAME: )
        CHECK-LLVMLITE-LE44: !DILocalVariable(
        CHECK-LLVMLITE-LE44-SAME: name: "a"
        CHECK-LLVMLITE-LE44-SAME: type: [[DBG103]]
        CHECK-LLVMLITE-LE44-SAME: )

        CHECK-LLVMLITE-GE45: [[DBG89:.+]] = !DIBasicType(encoding: DW_ATE_float, name: "float", size: 32)
        CHECK-LLVMLITE-GE45: distinct !DISubprogram
        CHECK-LLVMLITE-GE45: [[DBG97:.+]] = !DIDerivedType(
        CHECK-LLVMLITE-GE45-SAME: baseType: [[DBG89]]
        CHECK-LLVMLITE-GE45-SAME: name: "real"
        CHECK-LLVMLITE-GE45-SAME: offset: 0
        CHECK-LLVMLITE-GE45-SAME: size: 32
        CHECK-LLVMLITE-GE45-SAME: tag: DW_TAG_member
        CHECK-LLVMLITE-GE45-SAME: )
        CHECK-LLVMLITE-GE45: [[DBG98:.+]] = !DIDerivedType(
        CHECK-LLVMLITE-GE45-SAME: baseType: [[DBG89]]
        CHECK-LLVMLITE-GE45-SAME: name: "imag"
        CHECK-LLVMLITE-GE45-SAME: offset: 32
        CHECK-LLVMLITE-GE45-SAME: size: 32
        CHECK-LLVMLITE-GE45-SAME: tag: DW_TAG_member
        CHECK-LLVMLITE-GE45-SAME: )
        CHECK-LLVMLITE-GE45: [[DBG99:.+]] = !{ [[DBG97]], [[DBG98]] }
        CHECK-LLVMLITE-GE45: [[DBG100:.+]] = distinct !DICompositeType(
        CHECK-LLVMLITE-GE45-SAME: elements: [[DBG99]]
        CHECK-LLVMLITE-GE45-SAME: identifier: "{float, float}"
        CHECK-LLVMLITE-GE45-SAME: name: "complex64 ({float, float})"
        CHECK-LLVMLITE-GE45-SAME: size: 64
        CHECK-LLVMLITE-GE45-SAME: tag: DW_TAG_structure_type
        CHECK-LLVMLITE-GE45-SAME: )
        CHECK-LLVMLITE-GE45: !DILocalVariable(
        CHECK-LLVMLITE-GE45-SAME: name: "a"
        CHECK-LLVMLITE-GE45-SAME: type: [[DBG100]]
        CHECK-LLVMLITE-GE45-SAME: )
        """,
    ),
    (
        types.complex128,
        """
        CHECK-LLVMLITE-LE44: distinct !DISubprogram
        CHECK-LLVMLITE-LE44: [[DBG98:.+]] = !DIBasicType(encoding: DW_ATE_float, name: "double", size: 64)
        CHECK-LLVMLITE-LE44: [[DBG99:.+]] = !DIDerivedType(
        CHECK-LLVMLITE-LE44-SAME: baseType: [[DBG98]]
        CHECK-LLVMLITE-LE44-SAME: name: "real"
        CHECK-LLVMLITE-LE44-SAME: offset: 0
        CHECK-LLVMLITE-LE44-SAME: size: 64
        CHECK-LLVMLITE-LE44-SAME: tag: DW_TAG_member
        CHECK-LLVMLITE-LE44-SAME: )
        CHECK-LLVMLITE-LE44: [[DBG100:.+]] = !DIBasicType(encoding: DW_ATE_float, name: "double", size: 64)
        CHECK-LLVMLITE-LE44: [[DBG101:.+]] = !DIDerivedType(
        CHECK-LLVMLITE-LE44-SAME: baseType: [[DBG100]]
        CHECK-LLVMLITE-LE44-SAME: name: "imag"
        CHECK-LLVMLITE-LE44-SAME: offset: 64
        CHECK-LLVMLITE-LE44-SAME: size: 64
        CHECK-LLVMLITE-LE44-SAME: tag: DW_TAG_member
        CHECK-LLVMLITE-LE44-SAME: )
        CHECK-LLVMLITE-LE44: [[DBG102:.+]] = !{ [[DBG99]], [[DBG101]] }
        CHECK-LLVMLITE-LE44: [[DBG103:.+]] = distinct !DICompositeType(
        CHECK-LLVMLITE-LE44-SAME: elements: [[DBG102]]
        CHECK-LLVMLITE-LE44-SAME: identifier: "{double, double}"
        CHECK-LLVMLITE-LE44-SAME: name: "complex128 ({double, double})"
        CHECK-LLVMLITE-LE44-SAME: size: 128
        CHECK-LLVMLITE-LE44-SAME: tag: DW_TAG_structure_type
        CHECK-LLVMLITE-LE44-SAME: )
        CHECK-LLVMLITE-LE44: !DILocalVariable(
        CHECK-LLVMLITE-LE44-SAME: name: "a"
        CHECK-LLVMLITE-LE44-SAME: type: [[DBG103]]
        CHECK-LLVMLITE-LE44-SAME: )

        CHECK-LLVMLITE-GE45: [[DBG89:.+]] = !DIBasicType(encoding: DW_ATE_float, name: "double", size: 64)
        CHECK-LLVMLITE-GE45: !DISubprogram
        CHECK-LLVMLITE-GE45: [[DBG97:.+]] = !DIDerivedType(
        CHECK-LLVMLITE-GE45-SAME: baseType: [[DBG89]]
        CHECK-LLVMLITE-GE45-SAME: name: "real"
        CHECK-LLVMLITE-GE45-SAME: offset: 0
        CHECK-LLVMLITE-GE45-SAME: size: 64
        CHECK-LLVMLITE-GE45-SAME: tag: DW_TAG_member
        CHECK-LLVMLITE-GE45-SAME: )
        CHECK-LLVMLITE-GE45: [[DBG98:.+]] = !DIDerivedType(
        CHECK-LLVMLITE-GE45-SAME: baseType: [[DBG89]]
        CHECK-LLVMLITE-GE45-SAME: name: "imag"
        CHECK-LLVMLITE-GE45-SAME: offset: 64
        CHECK-LLVMLITE-GE45-SAME: size: 64
        CHECK-LLVMLITE-GE45-SAME: tag: DW_TAG_member
        CHECK-LLVMLITE-GE45-SAME: )
        CHECK-LLVMLITE-GE45: [[DBG99:.+]] = !{ [[DBG97]], [[DBG98]] }
        CHECK-LLVMLITE-GE45: [[DBG100:.+]] = distinct !DICompositeType(
        CHECK-LLVMLITE-GE45-SAME: elements: [[DBG99]]
        CHECK-LLVMLITE-GE45-SAME: identifier: "{double, double}"
        CHECK-LLVMLITE-GE45-SAME: name: "complex128 ({double, double})"
        CHECK-LLVMLITE-GE45-SAME: size: 128
        CHECK-LLVMLITE-GE45-SAME: tag: DW_TAG_structure_type
        CHECK-LLVMLITE-GE45-SAME: )
        CHECK-LLVMLITE-GE45: !DILocalVariable(
        CHECK-LLVMLITE-GE45-SAME: name: "a"
        CHECK-LLVMLITE-GE45-SAME: type: [[DBG100]]
        CHECK-LLVMLITE-GE45-SAME: )
        """,
    ),
)

array_types: tuple[tuple[types.Type, str]] = (
    (
        types.float32[::1],
        """
        CHECK: distinct !DICompileUnit
        CHECK: distinct !DISubprogram

        CHECK-LLVMLITE-LE44: [[DBG127:.+]] = !DIBasicType(encoding: DW_ATE_boolean, name: "i8", size: 8)
        CHECK-LLVMLITE-LE44: [[DBG128:.+]] = !DIDerivedType(
        CHECK-LLVMLITE-LE44-SAME: baseType: [[DBG127]]
        CHECK-LLVMLITE-LE44-SAME: size: 64
        CHECK-LLVMLITE-LE44-SAME: tag: DW_TAG_pointer_type
        CHECK-LLVMLITE-LE44-SAME: )
        CHECK-LLVMLITE-LE44: [[DBG129:.+]] = !DIDerivedType(
        CHECK-LLVMLITE-LE44-SAME: baseType: [[DBG128]]
        CHECK-LLVMLITE-LE44-SAME: name: "meminfo"
        CHECK-LLVMLITE-LE44-SAME: offset: 0
        CHECK-LLVMLITE-LE44-SAME: size: 64
        CHECK-LLVMLITE-LE44-SAME: tag: DW_TAG_member
        CHECK-LLVMLITE-LE44-SAME: )
        CHECK-LLVMLITE-LE44: [[DBG130:.+]] = !DIBasicType(encoding: DW_ATE_boolean, name: "i8", size: 8)
        CHECK-LLVMLITE-LE44: [[DBG131:.+]] = !DIDerivedType(
        CHECK-LLVMLITE-LE44-SAME: baseType: [[DBG130]]
        CHECK-LLVMLITE-LE44-SAME: size: 64
        CHECK-LLVMLITE-LE44-SAME: tag: DW_TAG_pointer_type
        CHECK-LLVMLITE-LE44-SAME: )
        CHECK-LLVMLITE-LE44: [[DBG132:.+]] = !DIDerivedType(
        CHECK-LLVMLITE-LE44-SAME: baseType: [[DBG131]]
        CHECK-LLVMLITE-LE44-SAME: name: "parent"
        CHECK-LLVMLITE-LE44-SAME: offset: 64
        CHECK-LLVMLITE-LE44-SAME: size: 64
        CHECK-LLVMLITE-LE44-SAME: tag: DW_TAG_member
        CHECK-LLVMLITE-LE44-SAME: )
        CHECK-LLVMLITE-LE44: [[DBG133:.+]] = !DIBasicType(encoding: DW_ATE_signed, name: "int64", size: 64)
        CHECK-LLVMLITE-LE44: [[DBG134:.+]] = !DIDerivedType(
        CHECK-LLVMLITE-LE44-SAME: baseType: [[DBG133]]
        CHECK-LLVMLITE-LE44-SAME: name: "nitems"
        CHECK-LLVMLITE-LE44-SAME: offset: 128
        CHECK-LLVMLITE-LE44-SAME: size: 64
        CHECK-LLVMLITE-LE44-SAME: tag: DW_TAG_member
        CHECK-LLVMLITE-LE44-SAME: )
        CHECK-LLVMLITE-LE44: [[DBG135:.+]] = !DIBasicType(encoding: DW_ATE_signed, name: "int64", size: 64)
        CHECK-LLVMLITE-LE44: [[DBG136:.+]] = !DIDerivedType(
        CHECK-LLVMLITE-LE44-SAME: baseType: [[DBG135]]
        CHECK-LLVMLITE-LE44-SAME: name: "itemsize"
        CHECK-LLVMLITE-LE44-SAME: offset: 192
        CHECK-LLVMLITE-LE44-SAME: size: 64
        CHECK-LLVMLITE-LE44-SAME: tag: DW_TAG_member
        CHECK-LLVMLITE-LE44-SAME: )
        CHECK-LLVMLITE-LE44: [[DBG137:.+]] = !DIBasicType(encoding: DW_ATE_float, name: "float32", size: 32)
        CHECK-LLVMLITE-LE44: [[DBG138:.+]] = !DIDerivedType(
        CHECK-LLVMLITE-LE44-SAME: baseType: [[DBG137]]
        CHECK-LLVMLITE-LE44-SAME: size: 64
        CHECK-LLVMLITE-LE44-SAME: tag: DW_TAG_pointer_type
        CHECK-LLVMLITE-LE44-SAME: )
        CHECK-LLVMLITE-LE44: [[DBG139:.+]] = !DIDerivedType(
        CHECK-LLVMLITE-LE44-SAME: baseType: [[DBG138]]
        CHECK-LLVMLITE-LE44-SAME: name: "data"
        CHECK-LLVMLITE-LE44-SAME: offset: 256
        CHECK-LLVMLITE-LE44-SAME: size: 64
        CHECK-LLVMLITE-LE44-SAME: tag: DW_TAG_member
        CHECK-LLVMLITE-LE44-SAME: )
        CHECK-LLVMLITE-LE44: [[DBG140:.+]] = !DIBasicType(encoding: DW_ATE_unsigned, name: "i64", size: 64)
        CHECK-LLVMLITE-LE44: [[DBG141:.+]] = !DICompositeType(
        CHECK-LLVMLITE-LE44-SAME: baseType: [[DBG140]]
        CHECK-LLVMLITE-LE44-SAME: identifier: "[1 x i64]"
        CHECK-LLVMLITE-LE44-SAME: name: "UniTuple(int64 x 1) ([1 x i64])"
        CHECK-LLVMLITE-LE44-SAME: tag: DW_TAG_array_type
        CHECK-LLVMLITE-LE44-SAME: )
        CHECK-LLVMLITE-LE44: [[DBG142:.+]] = !DIDerivedType(
        CHECK-LLVMLITE-LE44-SAME: baseType: [[DBG141]]
        CHECK-LLVMLITE-LE44-SAME: name: "shape"
        CHECK-LLVMLITE-LE44-SAME: offset: 320
        CHECK-LLVMLITE-LE44-SAME: size: 64
        CHECK-LLVMLITE-LE44-SAME: tag: DW_TAG_member
        CHECK-LLVMLITE-LE44-SAME: )
        CHECK-LLVMLITE-LE44: [[DBG143:.+]] = !DIBasicType(encoding: DW_ATE_unsigned, name: "i64", size: 64)
        CHECK-LLVMLITE-LE44: [[DBG144:.+]] = !DICompositeType(
        CHECK-LLVMLITE-LE44-SAME: baseType: [[DBG143]]
        CHECK-LLVMLITE-LE44-SAME: identifier: "[1 x i64]"
        CHECK-LLVMLITE-LE44-SAME: name: "UniTuple(int64 x 1) ([1 x i64])"
        CHECK-LLVMLITE-LE44-SAME: size: 64
        CHECK-LLVMLITE-LE44-SAME: tag: DW_TAG_array_type
        CHECK-LLVMLITE-LE44-SAME: )
        CHECK-LLVMLITE-LE44: [[DBG145:.+]] = !DIDerivedType(
        CHECK-LLVMLITE-LE44-SAME: baseType: [[DBG144]]
        CHECK-LLVMLITE-LE44-SAME: name: "strides"
        CHECK-LLVMLITE-LE44-SAME: offset: 384
        CHECK-LLVMLITE-LE44-SAME: size: 64
        CHECK-LLVMLITE-LE44-SAME: tag: DW_TAG_member
        CHECK-LLVMLITE-LE44-SAME: )
        CHECK-LLVMLITE-LE44: [[DBG146:.+]] = !{ [[DBG129]], [[DBG132]], [[DBG134]], [[DBG136]], [[DBG139]], [[DBG142]], [[DBG145]] }
        CHECK-LLVMLITE-LE44: [[DBG147:.+]] = distinct !DICompositeType(
        CHECK-LLVMLITE-LE44-SAME: elements: [[DBG146]]
        CHECK-LLVMLITE-LE44-SAME: identifier: "{i8*, i8*, i64, i64, float*, [1 x i64], [1 x i64]}"
        CHECK-LLVMLITE-LE44-SAME: name: "array(float32, 1d, C) ({i8*, i8*, i64, i64, float*, [1 x i64], [1 x i64]})"
        CHECK-LLVMLITE-LE44-SAME: size: 448
        CHECK-LLVMLITE-LE44-SAME: tag: DW_TAG_structure_type
        CHECK-LLVMLITE-LE44-SAME: )
        CHECK-LLVMLITE-LE44: !DILocalVariable(
        CHECK-LLVMLITE-LE44-SAME: name: "a"
        CHECK-LLVMLITE-LE44-SAME: type: [[DBG147]]
        CHECK-LLVMLITE-LE44-SAME: )

        CHECK-LLVMLITE-GE45: [[DBG123:.+]] = !DIDerivedType(
        CHECK-LLVMLITE-GE45-SAME: baseType: [[DBG98:![0-9]+]]
        CHECK-LLVMLITE-GE45-SAME: size: 64
        CHECK-LLVMLITE-GE45-SAME: tag: DW_TAG_pointer_type
        CHECK-LLVMLITE-GE45-SAME: )
        CHECK-LLVMLITE-GE45: [[DBG124:.+]] = !DIDerivedType(
        CHECK-LLVMLITE-GE45-SAME: baseType: [[DBG123]]
        CHECK-LLVMLITE-GE45-SAME: name: "meminfo"
        CHECK-LLVMLITE-GE45-SAME: offset: 0
        CHECK-LLVMLITE-GE45-SAME: size: 64
        CHECK-LLVMLITE-GE45-SAME: tag: DW_TAG_member
        CHECK-LLVMLITE-GE45-SAME: )
        CHECK-LLVMLITE-GE45: [[DBG125:.+]] = !DIDerivedType(
        CHECK-LLVMLITE-GE45-SAME: baseType: [[DBG98]]
        CHECK-LLVMLITE-GE45-SAME: size: 64
        CHECK-LLVMLITE-GE45-SAME: tag: DW_TAG_pointer_type
        CHECK-LLVMLITE-GE45-SAME: )
        CHECK-LLVMLITE-GE45: [[DBG126:.+]] = !DIDerivedType(
        CHECK-LLVMLITE-GE45-SAME: baseType: [[DBG125]]
        CHECK-LLVMLITE-GE45-SAME: name: "parent"
        CHECK-LLVMLITE-GE45-SAME: offset: 64
        CHECK-LLVMLITE-GE45-SAME: size: 64
        CHECK-LLVMLITE-GE45-SAME: tag: DW_TAG_member
        CHECK-LLVMLITE-GE45-SAME: )
        CHECK-LLVMLITE-GE45: [[DBG127:.+]] = !DIDerivedType(
        CHECK-LLVMLITE-GE45-SAME: baseType: [[DBG105:![0-9]+]]
        CHECK-LLVMLITE-GE45-SAME: name: "nitems"
        CHECK-LLVMLITE-GE45-SAME: offset: 128
        CHECK-LLVMLITE-GE45-SAME: size: 64
        CHECK-LLVMLITE-GE45-SAME: tag: DW_TAG_member
        CHECK-LLVMLITE-GE45-SAME: )
        CHECK-LLVMLITE-GE45: [[DBG128:.+]] = !DIDerivedType(
        CHECK-LLVMLITE-GE45-SAME: baseType: [[DBG105]]
        CHECK-LLVMLITE-GE45-SAME: name: "itemsize"
        CHECK-LLVMLITE-GE45-SAME: offset: 192
        CHECK-LLVMLITE-GE45-SAME: size: 64
        CHECK-LLVMLITE-GE45-SAME: tag: DW_TAG_member
        CHECK-LLVMLITE-GE45-SAME: )
        CHECK-LLVMLITE-GE45: [[DBG129:.+]] = !DIDerivedType(
        CHECK-LLVMLITE-GE45-SAME: baseType: [[DBG108:![0-9]+]]
        CHECK-LLVMLITE-GE45-SAME: size: 64
        CHECK-LLVMLITE-GE45-SAME: tag: DW_TAG_pointer_type
        CHECK-LLVMLITE-GE45-SAME: )
        CHECK-LLVMLITE-GE45: [[DBG130:.+]] = !DIDerivedType(
        CHECK-LLVMLITE-GE45-SAME: baseType: [[DBG129]]
        CHECK-LLVMLITE-GE45-SAME: name: "data"
        CHECK-LLVMLITE-GE45-SAME: offset: 256
        CHECK-LLVMLITE-GE45-SAME: size: 64
        CHECK-LLVMLITE-GE45-SAME: tag: DW_TAG_member
        CHECK-LLVMLITE-GE45-SAME: )
        CHECK-LLVMLITE-GE45: [[DBG131:.+]] = !DICompositeType(
        CHECK-LLVMLITE-GE45-SAME: baseType: [[DBG111:![0-9]+]]
        CHECK-LLVMLITE-GE45-SAME: elements: [[DBG113:![0-9]+]]
        CHECK-LLVMLITE-GE45-SAME: identifier: "[1 x i64]"
        CHECK-LLVMLITE-GE45-SAME: name: "UniTuple(int64 x 1) ([1 x i64])"
        CHECK-LLVMLITE-GE45-SAME: size: 64
        CHECK-LLVMLITE-GE45-SAME: tag: DW_TAG_array_type
        CHECK-LLVMLITE-GE45-SAME: )
        CHECK-LLVMLITE-GE45: [[DBG132:.+]] = !DIDerivedType(
        CHECK-LLVMLITE-GE45-SAME: baseType: [[DBG131]]
        CHECK-LLVMLITE-GE45-SAME: name: "shape"
        CHECK-LLVMLITE-GE45-SAME: offset: 320
        CHECK-LLVMLITE-GE45-SAME: size: 64
        CHECK-LLVMLITE-GE45-SAME: tag: DW_TAG_member
        CHECK-LLVMLITE-GE45-SAME: )
        CHECK-LLVMLITE-GE45: [[DBG133:.+]] = !DICompositeType(
        CHECK-LLVMLITE-GE45-SAME: baseType: [[DBG111]]
        CHECK-LLVMLITE-GE45-SAME: elements: [[DBG113]]
        CHECK-LLVMLITE-GE45-SAME: identifier: "[1 x i64]"
        CHECK-LLVMLITE-GE45-SAME: name: "UniTuple(int64 x 1) ([1 x i64])"
        CHECK-LLVMLITE-GE45-SAME: size: 64
        CHECK-LLVMLITE-GE45-SAME: tag: DW_TAG_array_type
        CHECK-LLVMLITE-GE45-SAME: )
        CHECK-LLVMLITE-GE45: [[DBG134:.+]] = !DIDerivedType(
        CHECK-LLVMLITE-GE45-SAME: baseType: [[DBG133]]
        CHECK-LLVMLITE-GE45-SAME: name: "strides"
        CHECK-LLVMLITE-GE45-SAME: offset: 384
        CHECK-LLVMLITE-GE45-SAME: size: 64
        CHECK-LLVMLITE-GE45-SAME: tag: DW_TAG_member
        CHECK-LLVMLITE-GE45-SAME: )
        CHECK-LLVMLITE-GE45: [[DBG135:.+]] = !{ [[DBG124]], [[DBG126]], [[DBG127]], [[DBG128]], [[DBG130]], [[DBG132]], [[DBG134]] }
        CHECK-LLVMLITE-GE45: [[DBG136:.+]] = distinct !DICompositeType(
        CHECK-LLVMLITE-GE45-SAME: elements: [[DBG135]]
        CHECK-LLVMLITE-GE45-SAME: identifier: "{i8*, i8*, i64, i64, float*, [1 x i64], [1 x i64]}"
        CHECK-LLVMLITE-GE45-SAME: name: "array(float32, 1d, C) ({i8*, i8*, i64, i64, float*, [1 x i64], [1 x i64]})"
        CHECK-LLVMLITE-GE45-SAME: size: 448
        CHECK-LLVMLITE-GE45-SAME: tag: DW_TAG_structure_type
        CHECK-LLVMLITE-GE45-SAME: )
        CHECK-LLVMLITE-GE45: [[DBG27:![0-9]+]] = !DILocalVariable(
        CHECK-LLVMLITE-GE45-SAME: name: "a"
        CHECK-LLVMLITE-GE45-SAME: type: [[DBG136]]
        CHECK-LLVMLITE-GE45-SAME: )
        """,
    ),
)


@skip_on_cudasim("Simulator does not produce debug dumps")
class TestCudaDebugInfoTypes(CUDATestCase):
    def test_ditypes(self):
        llvmlite_minor_version = int(llvmlite.__version__.split(".")[1])
        prefixes = (
            "CHECK",
            "CHECK-LLVMLITE-LE44"
            if llvmlite_minor_version <= 44
            else "CHECK-LLVMLITE-GE45",
        )

        def sanitize_name(name: str) -> str:
            return "".join(filter(lambda c: c.isalnum(), name))

        for numba_type, checks in scalar_types + array_types:
            with self.subTest(
                f"Test DITypes for {sanitize_name(numba_type.name)}"
            ):

                @cuda.jit((numba_type,), debug=True)
                def foo(a):
                    pass

                ir = foo.inspect_llvm()[foo.signatures[0]]
                self.assertFileCheckMatches(ir, checks, check_prefixes=prefixes)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from numba import cuda
from numba.cuda.testing import CUDATestCase, skip_on_cudasim, skip_unless_cc_53


class TestIsFP16Supported(CUDATestCase):
    def test_is_fp16_supported(self):
        self.assertTrue(cuda.is_float16_supported())

    @skip_on_cudasim
    @skip_unless_cc_53
    def test_device_supports_float16(self):
        self.assertTrue(cuda.get_current_device().supports_float16)



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import pathlib

from setuptools import setup
from setuptools.command.build_py import build_py
from setuptools.command.editable_wheel import editable_wheel, _TopLevelFinder

REDIRECTOR_PTH = "_numba_cuda_redirector.pth"
REDIRECTOR_PY = "_numba_cuda_redirector.py"
SITE_PACKAGES = pathlib.Path("site-packages")


# Adapted from https://stackoverflow.com/a/71137790
class build_py_with_redirector(build_py):  # noqa: N801
    """Include the redirector files in the generated wheel."""

    def copy_redirector_file(self, source, destination="."):
        destination = pathlib.Path(self.build_lib) / destination
        self.copy_file(str(source), str(destination), preserve_mode=0)

    def run(self):
        super().run()
        self.copy_redirector_file(SITE_PACKAGES / REDIRECTOR_PTH)
        self.copy_redirector_file(SITE_PACKAGES / REDIRECTOR_PY)

    def get_source_files(self):
        src = super().get_source_files()
        src.extend(
            [
                str(SITE_PACKAGES / REDIRECTOR_PTH),
                str(SITE_PACKAGES / REDIRECTOR_PY),
            ]
        )
        return src

    def get_output_mapping(self):
        mapping = super().get_output_mapping()
        build_lib = pathlib.Path(self.build_lib)
        mapping[str(build_lib / REDIRECTOR_PTH)] = REDIRECTOR_PTH
        mapping[str(build_lib / REDIRECTOR_PY)] = REDIRECTOR_PY
        return mapping


class TopLevelFinderWithRedirector(_TopLevelFinder):
    """Include the redirector files in the editable wheel."""

    def get_implementation(self):
        for item in super().get_implementation():
            yield item

        with open(SITE_PACKAGES / REDIRECTOR_PTH) as f:
            yield (REDIRECTOR_PTH, f.read())

        with open(SITE_PACKAGES / REDIRECTOR_PY) as f:
            yield (REDIRECTOR_PY, f.read())


class editable_wheel_with_redirector(editable_wheel):
    def _select_strategy(self, name, tag, build_lib):
        # The default mode is "lenient" - others are "strict" and "compat".
        # "compat" is deprecated. "strict" creates a tree of links to files in
        # the repo. It could be implemented, but we only handle the default
        # case for now.
        if self.mode is not None and self.mode != "lenient":
            raise RuntimeError(
                "Only lenient mode is supported for editable "
                f"install. Current mode is {self.mode}"
            )

        return TopLevelFinderWithRedirector(self.distribution, name)


setup(
    cmdclass={
        "build_py": build_py_with_redirector,
        "editable_wheel": editable_wheel_with_redirector,
    }
)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""
Separate module with function samples for serialization tests,
to avoid issues with __main__.
"""

import math
from math import sqrt
import numpy as np
import numpy.random as nprand

from numba import jit
from numba.core import types


@jit((types.int32, types.int32))
def add_with_sig(a, b):
    return a + b


@jit
def add_without_sig(a, b):
    return a + b


@jit(nopython=True)
def add_nopython(a, b):
    return a + b


@jit(nopython=True)
def add_nopython_fail(a, b):
    object()
    return a + b


def closure(a):
    @jit(nopython=True)
    def inner(b, c):
        return a + b + c

    return inner


K = 3.0


def closure_with_globals(x, **jit_args):
    @jit(**jit_args)
    def inner(y):
        # Exercise a builtin function and a module-level constant
        k = max(K, K + 1)
        # Exercise two functions from another module, one accessed with
        # dotted notation, one imported explicitly.
        return math.hypot(x, y) + sqrt(k)

    return inner


@jit(nopython=True)
def other_function(x, y):
    return math.hypot(x, y)


@jit(forceobj=True)
def get_global_objmode(x):
    return K * x


@jit(nopython=True)
def get_renamed_module(x):
    nprand.seed(42)
    return np.cos(x), nprand.random()


def closure_calling_other_function(x):
    @jit(nopython=True)
    def inner(y, z):
        return other_function(x, y) + z

    return inner


def closure_calling_other_closure(x):
    @jit(nopython=True)
    def other_inner(y):
        return math.hypot(x, y)

    @jit(nopython=True)
    def inner(y):
        return other_inner(y) + x

    return inner


# A dynamic function calling a builtin function
def _get_dyn_func(**jit_args):
    code = """
        def dyn_func(x):
            res = 0
            for i in range(x):
                res += x
            return res
        """
    ns = {}
    exec(code.strip(), ns)
    return jit(**jit_args)(ns["dyn_func"])


dyn_func = _get_dyn_func(nopython=True)
dyn_func_objmode = _get_dyn_func(forceobj=True)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from abc import abstractmethod, ABCMeta
from numba.misc.llvm_pass_timings import PassTimingsCollection


class CodeLibrary(metaclass=ABCMeta):
    """
    An interface for bundling LLVM code together and compiling it.
    It is tied to a *codegen* instance (e.g. JITCUDACodegen) that will
    determine how the LLVM code is transformed and linked together.
    """

    _finalized = False
    _object_caching_enabled = False
    _disable_inspection = False

    def __init__(self, codegen: "Codegen", name: str):
        self._codegen = codegen
        self._name = name
        ptc_name = f"{self.__class__.__name__}({self._name!r})"
        self._recorded_timings = PassTimingsCollection(ptc_name)
        # Track names of the dynamic globals
        self._dynamic_globals = []

    @property
    def has_dynamic_globals(self):
        self._ensure_finalized()
        return len(self._dynamic_globals) > 0

    @property
    def recorded_timings(self):
        return self._recorded_timings

    @property
    def codegen(self):
        """
        The codegen object owning this library.
        """
        return self._codegen

    @property
    def name(self):
        return self._name

    def __repr__(self):
        return "<Library %r at 0x%x>" % (self.name, id(self))

    def _raise_if_finalized(self):
        if self._finalized:
            raise RuntimeError(
                "operation impossible on finalized object %r" % (self,)
            )

    def _ensure_finalized(self):
        if not self._finalized:
            self.finalize()

    def create_ir_module(self, name):
        """
        Create an LLVM IR module for use by this library.
        """
        self._raise_if_finalized()
        ir_module = self._codegen._create_empty_module(name)
        return ir_module

    @abstractmethod
    def add_linking_library(self, library):
        """
        Add a library for linking into this library, without losing
        the original library.
        """

    @abstractmethod
    def add_ir_module(self, ir_module):
        """
        Add an LLVM IR module's contents to this library.
        """

    @abstractmethod
    def finalize(self):
        """
        Finalize the library.  After this call, nothing can be added anymore.
        Finalization involves various stages of code optimization and
        linking.
        """

    @abstractmethod
    def get_function(self, name):
        """
        Return the function named ``name``.
        """

    @abstractmethod
    def get_llvm_str(self):
        """
        Get the human-readable form of the LLVM module.
        """

    @abstractmethod
    def get_asm_str(self):
        """
        Get the human-readable assembly.
        """

    #
    # Object cache hooks and serialization
    #

    def enable_object_caching(self):
        self._object_caching_enabled = True
        self._compiled_object = None
        self._compiled = False

    def _get_compiled_object(self):
        if not self._object_caching_enabled:
            raise ValueError("object caching not enabled in %s" % (self,))
        if self._compiled_object is None:
            raise RuntimeError("no compiled object yet for %s" % (self,))
        return self._compiled_object

    def _set_compiled_object(self, value):
        if not self._object_caching_enabled:
            raise ValueError("object caching not enabled in %s" % (self,))
        if self._compiled:
            raise ValueError("library already compiled: %s" % (self,))
        self._compiled_object = value
        self._disable_inspection = True


class Codegen(metaclass=ABCMeta):
    """
    Base Codegen class. It is expected that subclasses set the class attribute
    ``_library_class``, indicating the CodeLibrary class for the target.

    Subclasses should also initialize:

    ``self._data_layout``: the data layout for the target.
    ``self._target_data``: the binding layer ``TargetData`` for the target.
    """

    @abstractmethod
    def _create_empty_module(self, name):
        """
        Create a new empty module suitable for the target.
        """

    @abstractmethod
    def _add_module(self, module):
        """
        Add a module to the execution engine. Ownership of the module is
        transferred to the engine.
        """

    @property
    def target_data(self):
        """
        The LLVM "target data" object for this codegen instance.
        """
        return self._target_data

    def create_library(self, name, **kwargs):
        """
        Create a :class:`CodeLibrary` object for use with this codegen
        instance.
        """
        return self._library_class(self, name, **kwargs)

    def unserialize_library(self, serialized):
        return self._library_class._unserialize(self, serialized)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from numba.cuda import libdevice, libdevicefuncs
from numba.cuda.typing.templates import ConcreteTemplate, Registry

registry = Registry()
register_global = registry.register_global


def libdevice_declare(func, retty, args):
    class Libdevice_function(ConcreteTemplate):
        cases = [libdevicefuncs.create_signature(retty, args)]

    pyfunc = getattr(libdevice, func[5:])
    register_global(pyfunc)(Libdevice_function)


for func, (retty, args) in libdevicefuncs.functions.items():
    libdevice_declare(func, retty, args)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import numpy as np
from numba.cuda.cudadrv import driver
from numba import cuda
from numba.cuda.testing import unittest, ContextResettingTestCase


class TestHostAlloc(ContextResettingTestCase):
    def test_host_alloc_driver(self):
        n = 32
        mem = cuda.current_context().memhostalloc(n, mapped=True)

        dtype = np.dtype(np.uint8)
        ary = np.ndarray(shape=n // dtype.itemsize, dtype=dtype, buffer=mem)

        magic = 0xAB
        driver.device_memset(mem, magic, n)

        self.assertTrue(np.all(ary == magic))

        ary.fill(n)

        recv = np.empty_like(ary)

        driver.device_to_host(recv, mem, ary.size)

        self.assertTrue(np.all(ary == recv))
        self.assertTrue(np.all(recv == n))

    def test_host_alloc_pinned(self):
        ary = cuda.pinned_array(10, dtype=np.uint32)
        ary.fill(123)
        self.assertTrue(all(ary == 123))
        devary = cuda.to_device(ary)
        driver.device_memset(devary, 0, driver.device_memory_size(devary))
        self.assertTrue(all(ary == 123))
        devary.copy_to_host(ary)
        self.assertTrue(all(ary == 0))

    def test_host_alloc_mapped(self):
        ary = cuda.mapped_array(10, dtype=np.uint32)
        ary.fill(123)
        self.assertTrue(all(ary == 123))
        driver.device_memset(ary, 0, driver.device_memory_size(ary))
        self.assertTrue(all(ary == 0))
        self.assertTrue(sum(ary != 0) == 0)

    def test_host_operators(self):
        for ary in [
            cuda.mapped_array(10, dtype=np.uint32),
            cuda.pinned_array(10, dtype=np.uint32),
        ]:
            ary[:] = range(10)
            self.assertTrue(sum(ary + 1) == 55)
            self.assertTrue(sum((ary + 1) * 2 - 1) == 100)
            self.assertTrue(sum(ary < 5) == 5)
            self.assertTrue(sum(ary <= 5) == 6)
            self.assertTrue(sum(ary > 6) == 3)
            self.assertTrue(sum(ary >= 6) == 4)
            self.assertTrue(sum(ary**2) == 285)
            self.assertTrue(sum(ary // 2) == 20)
            self.assertTrue(sum(ary / 2.0) == 22.5)
            self.assertTrue(sum(ary % 2) == 5)


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

# CUDA built-in Vector Types
# https://docs.nvidia.com/cuda/cuda-c-programming-guide/index.html#built-in-vector-types

from typing import List, Tuple, Dict

from numba import types
from numba.cuda import cgutils
from numba.core.extending import models
from numba.core.imputils import Registry as ImplRegistry
from numba.cuda.typing.templates import ConcreteTemplate
from numba.cuda.typing.templates import Registry as TypingRegistry
from numba.cuda.typing.templates import signature
from numba.cuda import stubs
from numba.cuda.errors import CudaLoweringError
from numba.cuda.extending import make_attribute_wrapper, register_model


typing_registry = TypingRegistry()
impl_registry = ImplRegistry()

register = typing_registry.register
register_attr = typing_registry.register_attr
register_global = typing_registry.register_global
lower = impl_registry.lower


class VectorType(types.Type):
    def __init__(self, name, base_type, attr_names, user_facing_object):
        self._base_type = base_type
        self._attr_names = attr_names
        self._user_facing_object = user_facing_object
        super().__init__(name=name)

    @property
    def base_type(self):
        return self._base_type

    @property
    def attr_names(self):
        return self._attr_names

    @property
    def num_elements(self):
        return len(self._attr_names)

    @property
    def user_facing_object(self):
        return self._user_facing_object


def make_vector_type(
    name: str,
    base_type: types.Type,
    attr_names: Tuple[str, ...],
    user_facing_object,
) -> types.Type:
    """Create a vector type.

    Parameters
    ----------
    name: str
        The name of the type.
    base_type: numba.types.Type
        The primitive type for each element in the vector.
    attr_names: tuple of str
        Name for each attribute.
    user_facing_object: object
        The handle to be used in cuda kernel.
    """

    class _VectorType(VectorType):
        """Internal instantiation of VectorType."""

        pass

    class VectorTypeModel(models.StructModel):
        def __init__(self, dmm, fe_type):
            members = [(attr_name, base_type) for attr_name in attr_names]
            super().__init__(dmm, fe_type, members)

    vector_type = _VectorType(name, base_type, attr_names, user_facing_object)
    register_model(_VectorType)(VectorTypeModel)
    for attr_name in attr_names:
        make_attribute_wrapper(_VectorType, attr_name, attr_name)

    return vector_type


def enable_vector_type_ctor(
    vector_type: VectorType, overloads: List[List[types.Type]]
):
    """Create typing and lowering for vector type constructor.

    Parameters
    ----------
    vector_type: VectorType
        The type whose constructor to type and lower.
    overloads: List of argument types
        A list containing different overloads of the constructor. Each base type
        in the argument list should either be primitive type or VectorType.
    """
    ctor = vector_type.user_facing_object

    @register
    class CtorTemplate(ConcreteTemplate):
        key = ctor
        cases = [signature(vector_type, *arglist) for arglist in overloads]

    register_global(ctor, types.Function(CtorTemplate))

    # Lowering

    def make_lowering(fml_arg_list):
        """Meta function to create a lowering for the constructor. Flattens
        the arguments by converting vector_type into load instructions for each
        of its attributes. Such as float2 -> float2.x, float2.y.
        """

        def lowering(context, builder, sig, actual_args):
            # A list of elements to assign from
            source_list = []
            # Convert the list of argument types to a list of load IRs.
            for argidx, fml_arg in enumerate(fml_arg_list):
                if isinstance(fml_arg, VectorType):
                    pxy = cgutils.create_struct_proxy(fml_arg)(
                        context, builder, actual_args[argidx]
                    )
                    source_list += [
                        getattr(pxy, attr) for attr in fml_arg.attr_names
                    ]
                else:
                    # assumed primitive type
                    source_list.append(actual_args[argidx])

            if len(source_list) != vector_type.num_elements:
                raise CudaLoweringError(
                    f"Unmatched number of source elements ({len(source_list)}) "
                    "and target elements ({vector_type.num_elements})."
                )

            out = cgutils.create_struct_proxy(vector_type)(context, builder)

            for attr_name, source in zip(vector_type.attr_names, source_list):
                setattr(out, attr_name, source)
            return out._getvalue()

        return lowering

    for arglist in overloads:
        lowering = make_lowering(arglist)
        lower(ctor, *arglist)(lowering)


vector_types: Dict[str, VectorType] = {}


def build_constructor_overloads(base_type, vty_name, num_elements, arglists, l):
    """
    For a given vector type, build a list of overloads for its constructor.
    """

    # TODO: speed up with memoization
    if num_elements == 0:
        arglists.append(l[:])

    for i in range(1, num_elements + 1):
        if i == 1:
            # For 1-element component, it can construct with either a
            # primitive type or other 1-element component.
            l.append(base_type)
            build_constructor_overloads(
                base_type, vty_name, num_elements - i, arglists, l
            )
            l.pop(-1)

            l.append(vector_types[f"{vty_name[:-1]}1"])
            build_constructor_overloads(
                base_type, vty_name, num_elements - i, arglists, l
            )
            l.pop(-1)
        else:
            l.append(vector_types[f"{vty_name[:-1]}{i}"])
            build_constructor_overloads(
                base_type, vty_name, num_elements - i, arglists, l
            )
            l.pop(-1)


def _initialize():
    """
    Construct the vector types, populate `vector_types` dictionary, and
    enable the constructors.
    """
    vector_type_attribute_names = ("x", "y", "z", "w")
    for stub in stubs._vector_type_stubs:
        type_name = stub.__name__
        base_type = getattr(types, type_name[:-2])
        num_elements = int(type_name[-1])
        attributes = vector_type_attribute_names[:num_elements]
        vector_type = make_vector_type(type_name, base_type, attributes, stub)
        vector_types[type_name] = vector_type

    for vty in vector_types.values():
        arglists, l = [], []
        build_constructor_overloads(
            vty.base_type, vty.name, vty.num_elements, arglists, l
        )
        enable_vector_type_ctor(vty, arglists)


_initialize()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import unittest

from numba.cuda.testing import CUDATestCase, skip_on_cudasim
from numba.cuda.tests.support import captured_stdout


@skip_on_cudasim("cudasim doesn't support cuda import at non-top-level")
class TestUFunc(CUDATestCase):
    """
    Test calling a UFunc
    """

    def setUp(self):
        # Prevent output from this test showing
        # up when running the test suite
        self._captured_stdout = captured_stdout()
        self._captured_stdout.__enter__()
        super().setUp()

    def tearDown(self):
        # No exception type, value, or traceback
        self._captured_stdout.__exit__(None, None, None)
        super().tearDown()

    def test_ex_cuda_ufunc_call(self):
        # ex_cuda_ufunc.begin
        import numpy as np
        from numba import cuda

        # A kernel calling a ufunc (sin, in this case)
        @cuda.jit
        def f(r, x):
            # Compute sin(x) with result written to r
            np.sin(x, r)

        # Declare input and output arrays
        x = np.arange(10, dtype=np.float32) - 5
        r = np.zeros_like(x)

        # Launch kernel that calls the ufunc
        f[1, 1](r, x)

        # A quick sanity check demonstrating equality of the sine computed by
        # the sin ufunc inside the kernel, and NumPy's sin ufunc
        np.testing.assert_allclose(r, np.sin(x))
        # ex_cuda_ufunc.end


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import numpy as np


def prepare_shape_strides_dtype(shape, strides, dtype, order):
    dtype = np.dtype(dtype)
    if isinstance(shape, (float, np.floating)):
        raise TypeError("shape must be an integer or tuple of integers")
    if isinstance(shape, np.ndarray) and np.issubdtype(
        shape.dtype, np.floating
    ):
        raise TypeError("shape must be an integer or tuple of integers")
    if isinstance(shape, int):
        shape = (shape,)
    if isinstance(strides, int):
        strides = (strides,)
    else:
        strides = strides or _fill_stride_by_order(shape, dtype, order)
    return shape, strides, dtype


def _fill_stride_by_order(shape, dtype, order):
    nd = len(shape)
    if nd == 0:
        return ()
    strides = [0] * nd
    if order == "C":
        strides[-1] = dtype.itemsize
        for d in reversed(range(nd - 1)):
            strides[d] = strides[d + 1] * shape[d + 1]
    elif order == "F":
        strides[0] = dtype.itemsize
        for d in range(1, nd):
            strides[d] = strides[d - 1] * shape[d - 1]
    else:
        raise ValueError("must be either C/F order")
    return tuple(strides)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import numpy as np

from collections import namedtuple
from numba import void, int32, float32, float64
from numba import guvectorize
from numba import cuda
from numba.cuda.testing import skip_on_cudasim, CUDATestCase
import unittest
import warnings
from numba.core.errors import NumbaPerformanceWarning, TypingError
from numba.cuda.tests.support import override_config


def _get_matmulcore_gufunc(dtype=float32):
    @guvectorize(
        [void(dtype[:, :], dtype[:, :], dtype[:, :])],
        "(m,n),(n,p)->(m,p)",
        target="cuda",
    )
    def matmulcore(A, B, C):
        m, n = A.shape
        n, p = B.shape
        for i in range(m):
            for j in range(p):
                C[i, j] = 0
                for k in range(n):
                    C[i, j] += A[i, k] * B[k, j]

    return matmulcore


@skip_on_cudasim("ufunc API unsupported in the simulator")
class TestCUDAGufunc(CUDATestCase):
    def test_gufunc_small(self):
        gufunc = _get_matmulcore_gufunc()

        matrix_ct = 2
        A = np.arange(matrix_ct * 2 * 4, dtype=np.float32).reshape(
            matrix_ct, 2, 4
        )
        B = np.arange(matrix_ct * 4 * 5, dtype=np.float32).reshape(
            matrix_ct, 4, 5
        )

        C = gufunc(A, B)
        Gold = np.matmul(A, B)
        self.assertTrue(np.allclose(C, Gold))

    def test_gufunc_auto_transfer(self):
        gufunc = _get_matmulcore_gufunc()

        matrix_ct = 2
        A = np.arange(matrix_ct * 2 * 4, dtype=np.float32).reshape(
            matrix_ct, 2, 4
        )
        B = np.arange(matrix_ct * 4 * 5, dtype=np.float32).reshape(
            matrix_ct, 4, 5
        )

        dB = cuda.to_device(B)

        C = gufunc(A, dB).copy_to_host()
        Gold = np.matmul(A, B)
        self.assertTrue(np.allclose(C, Gold))

    def test_gufunc(self):
        gufunc = _get_matmulcore_gufunc()

        matrix_ct = 1001  # an odd number to test thread/block division in CUDA
        A = np.arange(matrix_ct * 2 * 4, dtype=np.float32).reshape(
            matrix_ct, 2, 4
        )
        B = np.arange(matrix_ct * 4 * 5, dtype=np.float32).reshape(
            matrix_ct, 4, 5
        )

        C = gufunc(A, B)
        Gold = np.matmul(A, B)
        self.assertTrue(np.allclose(C, Gold))

    def test_gufunc_hidim(self):
        gufunc = _get_matmulcore_gufunc()

        matrix_ct = 100  # an odd number to test thread/block division in CUDA
        A = np.arange(matrix_ct * 2 * 4, dtype=np.float32).reshape(4, 25, 2, 4)
        B = np.arange(matrix_ct * 4 * 5, dtype=np.float32).reshape(4, 25, 4, 5)

        C = gufunc(A, B)
        Gold = np.matmul(A, B)
        self.assertTrue(np.allclose(C, Gold))

    def test_gufunc_new_axis(self):
        gufunc = _get_matmulcore_gufunc(dtype=float64)

        X = np.random.randn(10, 3, 3)
        Y = np.random.randn(3, 3)

        gold = np.matmul(X, Y)

        res1 = gufunc(X, Y)
        np.testing.assert_allclose(gold, res1)

        res2 = gufunc(X, np.tile(Y, (10, 1, 1)))
        np.testing.assert_allclose(gold, res2)

    def test_gufunc_stream(self):
        gufunc = _get_matmulcore_gufunc()

        # cuda.driver.flush_pending_free()
        matrix_ct = 1001  # an odd number to test thread/block division in CUDA
        A = np.arange(matrix_ct * 2 * 4, dtype=np.float32).reshape(
            matrix_ct, 2, 4
        )
        B = np.arange(matrix_ct * 4 * 5, dtype=np.float32).reshape(
            matrix_ct, 4, 5
        )

        stream = cuda.stream()
        dA = cuda.to_device(A, stream)
        dB = cuda.to_device(B, stream)

        dC = cuda.device_array(shape=(1001, 2, 5), dtype=A.dtype, stream=stream)
        dC = gufunc(dA, dB, out=dC, stream=stream)
        C = dC.copy_to_host(stream=stream)
        stream.synchronize()

        Gold = np.matmul(A, B)

        self.assertTrue(np.allclose(C, Gold))

    def test_copy(self):
        @guvectorize([void(float32[:], float32[:])], "(x)->(x)", target="cuda")
        def copy(A, B):
            for i in range(B.size):
                B[i] = A[i]

        A = np.arange(10, dtype=np.float32) + 1
        B = np.zeros_like(A)
        copy(A, out=B)
        np.testing.assert_allclose(A, B)

    def test_copy_unspecified_return(self):
        # Ensure that behaviour is correct when the return type is not
        # specified in the signature.
        @guvectorize([(float32[:], float32[:])], "(x)->(x)", target="cuda")
        def copy(A, B):
            for i in range(B.size):
                B[i] = A[i]

        A = np.arange(10, dtype=np.float32) + 1
        B = np.zeros_like(A)
        copy(A, out=B)
        self.assertTrue(np.allclose(A, B))

    def test_copy_odd(self):
        @guvectorize([void(float32[:], float32[:])], "(x)->(x)", target="cuda")
        def copy(A, B):
            for i in range(B.size):
                B[i] = A[i]

        A = np.arange(11, dtype=np.float32) + 1
        B = np.zeros_like(A)
        copy(A, out=B)
        self.assertTrue(np.allclose(A, B))

    def test_copy2d(self):
        @guvectorize(
            [void(float32[:, :], float32[:, :])],
            "(x, y)->(x, y)",
            target="cuda",
        )
        def copy2d(A, B):
            for x in range(B.shape[0]):
                for y in range(B.shape[1]):
                    B[x, y] = A[x, y]

        A = np.arange(30, dtype=np.float32).reshape(5, 6) + 1
        B = np.zeros_like(A)
        copy2d(A, out=B)
        self.assertTrue(np.allclose(A, B))

    def test_not_supported_call_from_jit(self):
        # not supported
        @guvectorize([void(int32[:], int32[:])], "(n)->(n)", target="cuda")
        def gufunc_copy(A, b):
            for i in range(A.shape[0]):
                b[i] = A[i]

        @cuda.jit
        def cuda_jit(A, b):
            return gufunc_copy(A, b)

        A = np.arange(1024 * 32).astype("int32")
        b = np.zeros_like(A)
        msg = "Untyped global name 'gufunc_copy'.*"
        with self.assertRaisesRegex(TypingError, msg):
            cuda_jit[1, 1](A, b)

    # Test inefficient use of the GPU where the inputs are all mapped onto a
    # single thread in a single block.
    def test_inefficient_launch_configuration(self):
        @guvectorize(
            ["void(float32[:], float32[:], float32[:])"],
            "(n),(n)->(n)",
            target="cuda",
        )
        def numba_dist_cuda(a, b, dist):
            len = a.shape[0]
            for i in range(len):
                dist[i] = a[i] * b[i]

        a = np.random.rand(1024 * 32).astype("float32")
        b = np.random.rand(1024 * 32).astype("float32")
        dist = np.zeros(a.shape[0]).astype("float32")

        with override_config("CUDA_LOW_OCCUPANCY_WARNINGS", 1):
            with warnings.catch_warnings(record=True) as w:
                numba_dist_cuda(a, b, dist)
                self.assertEqual(w[0].category, NumbaPerformanceWarning)
                self.assertIn("Grid size", str(w[0].message))
                self.assertIn("low occupancy", str(w[0].message))

    def test_efficient_launch_configuration(self):
        @guvectorize(
            ["void(float32[:], float32[:], float32[:])"],
            "(n),(n)->(n)",
            nopython=True,
            target="cuda",
        )
        def numba_dist_cuda2(a, b, dist):
            len = a.shape[0]
            for i in range(len):
                dist[i] = a[i] * b[i]

        a = np.random.rand(524288 * 2).astype("float32").reshape((524288, 2))
        b = np.random.rand(524288 * 2).astype("float32").reshape((524288, 2))
        dist = np.zeros_like(a)

        with override_config("CUDA_LOW_OCCUPANCY_WARNINGS", 1):
            with warnings.catch_warnings(record=True) as w:
                numba_dist_cuda2(a, b, dist)
                self.assertEqual(len(w), 0)

    def test_nopython_flag(self):
        def foo(A, B):
            pass

        # nopython = True is fine
        guvectorize(
            [void(float32[:], float32[:])],
            "(x)->(x)",
            target="cuda",
            nopython=True,
        )(foo)

        # nopython = False is bad
        with self.assertRaises(TypeError) as raises:
            guvectorize(
                [void(float32[:], float32[:])],
                "(x)->(x)",
                target="cuda",
                nopython=False,
            )(foo)
        self.assertEqual("nopython flag must be True", str(raises.exception))

    def test_invalid_flags(self):
        # Check invalid flags
        def foo(A, B):
            pass

        with self.assertRaises(TypeError) as raises:
            guvectorize(
                [void(float32[:], float32[:])],
                "(x)->(x)",
                target="cuda",
                what1=True,
                ever2=False,
            )(foo)
        head = "The following target options are not supported:"
        msg = str(raises.exception)
        self.assertEqual(msg[: len(head)], head)
        items = msg[len(head) :].strip().split(",")
        items = [i.strip("'\" ") for i in items]
        self.assertEqual(set(["what1", "ever2"]), set(items))

    def test_duplicated_output(self):
        @guvectorize([void(float32[:], float32[:])], "(x)->(x)", target="cuda")
        def foo(inp, out):
            pass  # intentionally empty; never executed

        inp = out = np.zeros(10, dtype=np.float32)
        with self.assertRaises(ValueError) as raises:
            foo(inp, out, out=out)

        msg = "cannot specify argument 'out' as both positional and keyword"
        self.assertEqual(str(raises.exception), msg)

    def check_tuple_arg(self, a, b):
        @guvectorize(
            [(float64[:], float64[:], float64[:])], "(n),(n)->()", target="cuda"
        )
        def gu_reduce(x, y, r):
            s = 0
            for i in range(len(x)):
                s += x[i] * y[i]
            r[0] = s

        r = gu_reduce(a, b)
        expected = np.sum(np.asarray(a) * np.asarray(b), axis=1)
        np.testing.assert_equal(expected, r)

    def test_tuple_of_tuple_arg(self):
        a = ((1.0, 2.0, 3.0), (4.0, 5.0, 6.0))
        b = ((1.5, 2.5, 3.5), (4.5, 5.5, 6.5))
        self.check_tuple_arg(a, b)

    def test_tuple_of_namedtuple_arg(self):
        Point = namedtuple("Point", ("x", "y", "z"))
        a = (Point(x=1.0, y=2.0, z=3.0), Point(x=4.0, y=5.0, z=6.0))
        b = (Point(x=1.5, y=2.5, z=3.5), Point(x=4.5, y=5.5, z=6.5))
        self.check_tuple_arg(a, b)

    def test_tuple_of_array_arg(self):
        a = (np.asarray((1.0, 2.0, 3.0)), np.asarray((4.0, 5.0, 6.0)))
        b = (np.asarray((1.5, 2.5, 3.5)), np.asarray((4.5, 5.5, 6.5)))
        self.check_tuple_arg(a, b)

    def test_gufunc_name(self):
        gufunc = _get_matmulcore_gufunc()
        self.assertEqual(gufunc.__name__, "matmulcore")

    def test_bad_return_type(self):
        with self.assertRaises(TypeError) as te:

            @guvectorize([int32(int32[:], int32[:])], "(m)->(m)", target="cuda")
            def f(x, y):
                pass

        msg = str(te.exception)
        self.assertIn("guvectorized functions cannot return values", msg)
        self.assertIn("specifies int32 return type", msg)

    def test_incorrect_number_of_pos_args(self):
        @guvectorize(
            [(int32[:], int32[:], int32[:])], "(m),(m)->(m)", target="cuda"
        )
        def f(x, y, z):
            pass

        arr = np.arange(5)

        # Inputs only, too few
        with self.assertRaises(TypeError) as te:
            f(arr)

        msg = str(te.exception)
        self.assertIn("gufunc accepts 2 positional arguments", msg)
        self.assertIn("or 3 positional arguments", msg)
        self.assertIn("Got 1 positional argument.", msg)

        # Inputs and outputs, too many
        with self.assertRaises(TypeError) as te:
            f(arr, arr, arr, arr)

        msg = str(te.exception)
        self.assertIn("gufunc accepts 2 positional arguments", msg)
        self.assertIn("or 3 positional arguments", msg)
        self.assertIn("Got 4 positional arguments.", msg)


@skip_on_cudasim("ufunc API unsupported in the simulator")
class TestMultipleOutputs(CUDATestCase):
    def test_multiple_outputs_same_type_passed_in(self):
        @guvectorize(
            [void(float32[:], float32[:], float32[:])],
            "(x)->(x),(x)",
            target="cuda",
        )
        def copy(A, B, C):
            for i in range(B.size):
                B[i] = A[i]
                C[i] = A[i]

        A = np.arange(10, dtype=np.float32) + 1
        B = np.zeros_like(A)
        C = np.zeros_like(A)
        copy(A, B, C)
        np.testing.assert_allclose(A, B)
        np.testing.assert_allclose(A, C)

    def test_multiple_outputs_distinct_values(self):
        @guvectorize(
            [void(float32[:], float32[:], float32[:])],
            "(x)->(x),(x)",
            target="cuda",
        )
        def copy_and_double(A, B, C):
            for i in range(B.size):
                B[i] = A[i]
                C[i] = A[i] * 2

        A = np.arange(10, dtype=np.float32) + 1
        B = np.zeros_like(A)
        C = np.zeros_like(A)
        copy_and_double(A, B, C)
        np.testing.assert_allclose(A, B)
        np.testing.assert_allclose(A * 2, C)

    def test_multiple_output_allocation(self):
        @guvectorize(
            [void(float32[:], float32[:], float32[:])],
            "(x)->(x),(x)",
            target="cuda",
        )
        def copy_and_double(A, B, C):
            for i in range(B.size):
                B[i] = A[i]
                C[i] = A[i] * 2

        A = np.arange(10, dtype=np.float32) + 1
        B, C = copy_and_double(A)
        np.testing.assert_allclose(A, B)
        np.testing.assert_allclose(A * 2, C)

    def test_multiple_output_dtypes(self):
        @guvectorize(
            [void(int32[:], int32[:], float64[:])],
            "(x)->(x),(x)",
            target="cuda",
        )
        def copy_and_multiply(A, B, C):
            for i in range(B.size):
                B[i] = A[i]
                C[i] = A[i] * 1.5

        A = np.arange(10, dtype=np.int32) + 1
        B = np.zeros_like(A)
        C = np.zeros_like(A, dtype=np.float64)
        copy_and_multiply(A, B, C)
        np.testing.assert_allclose(A, B)
        np.testing.assert_allclose(A * np.float64(1.5), C)

    def test_incorrect_number_of_pos_args(self):
        @guvectorize(
            [(int32[:], int32[:], int32[:], int32[:])],
            "(m),(m)->(m),(m)",
            target="cuda",
        )
        def f(x, y, z, w):
            pass

        arr = np.arange(5)

        # Inputs only, too few
        with self.assertRaises(TypeError) as te:
            f(arr)

        msg = str(te.exception)
        self.assertIn("gufunc accepts 2 positional arguments", msg)
        self.assertIn("or 4 positional arguments", msg)
        self.assertIn("Got 1 positional argument.", msg)

        # Inputs and outputs, too many
        with self.assertRaises(TypeError) as te:
            f(arr, arr, arr, arr, arr)

        msg = str(te.exception)
        self.assertIn("gufunc accepts 2 positional arguments", msg)
        self.assertIn("or 4 positional arguments", msg)
        self.assertIn("Got 5 positional arguments.", msg)


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import numpy as np

from numba import cuda, complex64, int32, float64
from numba.cuda.testing import unittest, CUDATestCase
from numba.core.config import ENABLE_CUDASIM

CONST_EMPTY = np.array([])
CONST1D = np.arange(10, dtype=np.float64) / 2.0
CONST2D = np.asfortranarray(np.arange(100, dtype=np.int32).reshape(10, 10))
CONST3D = (np.arange(5 * 5 * 5, dtype=np.complex64).reshape(5, 5, 5) + 1j) / 2j
CONST3BYTES = np.arange(3, dtype=np.uint8)

CONST_RECORD_EMPTY = np.array([], dtype=[("x", float), ("y", int)])
CONST_RECORD = np.array([(1.0, 2), (3.0, 4)], dtype=[("x", float), ("y", int)])
CONST_RECORD_ALIGN = np.array(
    [(1, 2, 3, 0xDEADBEEF, 8), (4, 5, 6, 0xBEEFDEAD, 10)],
    dtype=np.dtype(
        dtype=[
            ("a", np.uint8),
            ("b", np.uint8),
            ("x", np.uint8),
            ("y", np.uint32),
            ("z", np.uint8),
        ],
        align=True,
    ),
)


def cuconstEmpty(A):
    C = cuda.const.array_like(CONST_EMPTY)
    i = cuda.grid(1)
    A[i] = len(C)


def cuconst(A):
    C = cuda.const.array_like(CONST1D)
    i = cuda.grid(1)

    # +1 or it'll be loaded & stored as a u32
    A[i] = C[i] + 1.0


def cuconst2d(A):
    C = cuda.const.array_like(CONST2D)
    i, j = cuda.grid(2)
    A[i, j] = C[i, j]


def cuconst3d(A):
    C = cuda.const.array_like(CONST3D)
    i = cuda.threadIdx.x
    j = cuda.threadIdx.y
    k = cuda.threadIdx.z
    A[i, j, k] = C[i, j, k]


def cuconstRecEmpty(A):
    C = cuda.const.array_like(CONST_RECORD_EMPTY)
    i = cuda.grid(1)
    A[i] = len(C)


def cuconstRec(A, B):
    C = cuda.const.array_like(CONST_RECORD)
    i = cuda.grid(1)
    A[i] = C[i]["x"]
    B[i] = C[i]["y"]


def cuconstRecAlign(A, B, C, D, E):
    Z = cuda.const.array_like(CONST_RECORD_ALIGN)
    i = cuda.grid(1)
    A[i] = Z[i]["a"]
    B[i] = Z[i]["b"]
    C[i] = Z[i]["x"]
    D[i] = Z[i]["y"]
    E[i] = Z[i]["z"]


def cuconstAlign(z):
    a = cuda.const.array_like(CONST3BYTES)
    b = cuda.const.array_like(CONST1D)
    i = cuda.grid(1)
    z[i] = a[i] + b[i]


class TestCudaConstantMemory(CUDATestCase):
    def test_const_array(self):
        sig = (float64[:],)
        jcuconst = cuda.jit(sig)(cuconst)
        A = np.zeros_like(CONST1D)
        jcuconst[2, 5](A)
        self.assertTrue(np.all(A == CONST1D + 1))

        if not ENABLE_CUDASIM:
            self.assertIn(
                "ld.const.f64",
                jcuconst.inspect_asm(sig),
                "as we're adding to it, load as a double",
            )

    def test_const_empty(self):
        jcuconstEmpty = cuda.jit("void(int64[:])")(cuconstEmpty)
        A = np.full(1, fill_value=-1, dtype=np.int64)
        jcuconstEmpty[1, 1](A)
        self.assertTrue(np.all(A == 0))

    def test_const_align(self):
        jcuconstAlign = cuda.jit("void(float64[:])")(cuconstAlign)
        A = np.full(3, fill_value=np.nan, dtype=float)
        jcuconstAlign[1, 3](A)
        self.assertTrue(np.all(A == (CONST3BYTES + CONST1D[:3])))

    def test_const_array_2d(self):
        sig = (int32[:, :],)
        jcuconst2d = cuda.jit(sig)(cuconst2d)
        A = np.zeros_like(CONST2D, order="C")
        jcuconst2d[(2, 2), (5, 5)](A)
        self.assertTrue(np.all(A == CONST2D))

        if not ENABLE_CUDASIM:
            self.assertIn(
                "ld.const.u32",
                jcuconst2d.inspect_asm(sig),
                "load the ints as ints",
            )

    def test_const_array_3d(self):
        sig = (complex64[:, :, :],)
        jcuconst3d = cuda.jit(sig)(cuconst3d)
        A = np.zeros_like(CONST3D, order="F")
        jcuconst3d[1, (5, 5, 5)](A)
        self.assertTrue(np.all(A == CONST3D))

        if not ENABLE_CUDASIM:
            asm = jcuconst3d.inspect_asm(sig)
            complex_load = "ld.const.v2.f32"
            description = "Load the complex as a vector of 2x f32"
            self.assertIn(complex_load, asm, description)

    def test_const_record_empty(self):
        jcuconstRecEmpty = cuda.jit("void(int64[:])")(cuconstRecEmpty)
        A = np.full(1, fill_value=-1, dtype=np.int64)
        jcuconstRecEmpty[1, 1](A)
        self.assertTrue(np.all(A == 0))

    def test_const_record(self):
        A = np.zeros(2, dtype=float)
        B = np.zeros(2, dtype=int)
        jcuconst = cuda.jit(cuconstRec).specialize(A, B)

        jcuconst[2, 1](A, B)
        np.testing.assert_allclose(A, CONST_RECORD["x"])
        np.testing.assert_allclose(B, CONST_RECORD["y"])

    def test_const_record_align(self):
        A = np.zeros(2, dtype=np.float64)
        B = np.zeros(2, dtype=np.float64)
        C = np.zeros(2, dtype=np.float64)
        D = np.zeros(2, dtype=np.float64)
        E = np.zeros(2, dtype=np.float64)
        jcuconst = cuda.jit(cuconstRecAlign).specialize(A, B, C, D, E)

        jcuconst[2, 1](A, B, C, D, E)
        np.testing.assert_allclose(A, CONST_RECORD_ALIGN["a"])
        np.testing.assert_allclose(B, CONST_RECORD_ALIGN["b"])
        np.testing.assert_allclose(C, CONST_RECORD_ALIGN["x"])
        np.testing.assert_allclose(D, CONST_RECORD_ALIGN["y"])
        np.testing.assert_allclose(E, CONST_RECORD_ALIGN["z"])


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from typing import List
from dataclasses import dataclass, field
from numba import cuda, float32
from numba.cuda.compiler import compile_ptx_for_current_device, compile_ptx
from math import cos, sin, tan, exp, log, log10, log2, pow, tanh
from operator import truediv
import numpy as np
from numba.cuda.testing import CUDATestCase, skip_on_cudasim, skip_unless_cc_75
import unittest


@dataclass
class FastMathCriterion:
    fast_expected: List[str] = field(default_factory=list)
    fast_unexpected: List[str] = field(default_factory=list)
    prec_expected: List[str] = field(default_factory=list)
    prec_unexpected: List[str] = field(default_factory=list)

    def check(self, test: CUDATestCase, fast: str, prec: str):
        test.assertTrue(all(i in fast for i in self.fast_expected))
        test.assertTrue(all(i not in fast for i in self.fast_unexpected))
        test.assertTrue(all(i in prec for i in self.prec_expected))
        test.assertTrue(all(i not in prec for i in self.prec_unexpected))


@skip_on_cudasim("Fastmath and PTX inspection not available on cudasim")
class TestFastMathOption(CUDATestCase):
    def _test_fast_math_common(self, pyfunc, sig, device, criterion):
        # Test jit code path
        fastver = cuda.jit(sig, device=device, fastmath=True)(pyfunc)
        precver = cuda.jit(sig, device=device)(pyfunc)

        criterion.check(
            self, fastver.inspect_asm(sig), precver.inspect_asm(sig)
        )

        # Test compile_ptx code path
        fastptx, _ = compile_ptx_for_current_device(
            pyfunc, sig, device=device, fastmath=True
        )
        precptx, _ = compile_ptx_for_current_device(pyfunc, sig, device=device)

        criterion.check(self, fastptx, precptx)

    def _test_fast_math_unary(self, op, criterion: FastMathCriterion):
        def kernel(r, x):
            r[0] = op(x)

        def device_function(x):
            return op(x)

        self._test_fast_math_common(
            kernel, (float32[::1], float32), device=False, criterion=criterion
        )
        self._test_fast_math_common(
            device_function, (float32,), device=True, criterion=criterion
        )

    def _test_fast_math_binary(self, op, criterion: FastMathCriterion):
        def kernel(r, x, y):
            r[0] = op(x, y)

        def device(x, y):
            return op(x, y)

        self._test_fast_math_common(
            kernel,
            (float32[::1], float32, float32),
            device=False,
            criterion=criterion,
        )
        self._test_fast_math_common(
            device, (float32, float32), device=True, criterion=criterion
        )

    def test_cosf(self):
        self._test_fast_math_unary(
            cos,
            FastMathCriterion(
                fast_expected=["cos.approx.ftz.f32 "],
                prec_unexpected=["cos.approx.ftz.f32 "],
            ),
        )

    def test_sinf(self):
        self._test_fast_math_unary(
            sin,
            FastMathCriterion(
                fast_expected=["sin.approx.ftz.f32 "],
                prec_unexpected=["sin.approx.ftz.f32 "],
            ),
        )

    def test_tanf(self):
        self._test_fast_math_unary(
            tan,
            FastMathCriterion(
                fast_expected=[
                    "sin.approx.ftz.f32 ",
                    "cos.approx.ftz.f32 ",
                    "div.approx.ftz.f32 ",
                ],
                prec_unexpected=["sin.approx.ftz.f32 "],
            ),
        )

    @skip_unless_cc_75
    def test_tanhf(self):
        self._test_fast_math_unary(
            tanh,
            FastMathCriterion(
                fast_expected=["tanh.approx.f32 "],
                prec_unexpected=["tanh.approx.f32 "],
            ),
        )

    def test_tanhf_compile_ptx(self):
        def tanh_kernel(r, x):
            r[0] = tanh(x)

        fastptx, _ = compile_ptx(
            tanh_kernel, (float32[::1], float32), fastmath=True
        )
        precptx, _ = compile_ptx(tanh_kernel, (float32[::1], float32))

        criterion = FastMathCriterion(
            fast_expected=["tanh.approx.f32 "],
            prec_unexpected=["tanh.approx.f32 "],
        )

        criterion.check(self, fastptx, precptx)

    def test_expf(self):
        self._test_fast_math_unary(
            exp,
            FastMathCriterion(
                fast_unexpected=["fma.rn.f32 "], prec_expected=["fma.rn.f32 "]
            ),
        )

    def test_logf(self):
        # Look for constant used to convert from log base 2 to log base e
        self._test_fast_math_unary(
            log,
            FastMathCriterion(
                fast_expected=["lg2.approx.ftz.f32 ", "0f3F317218"],
                prec_unexpected=["lg2.approx.ftz.f32 "],
            ),
        )

    def test_log10f(self):
        # Look for constant used to convert from log base 2 to log base 10
        self._test_fast_math_unary(
            log10,
            FastMathCriterion(
                fast_expected=["lg2.approx.ftz.f32 ", "0f3E9A209B"],
                prec_unexpected=["lg2.approx.ftz.f32 "],
            ),
        )

    def test_log2f(self):
        self._test_fast_math_unary(
            log2,
            FastMathCriterion(
                fast_expected=["lg2.approx.ftz.f32 "],
                prec_unexpected=["lg2.approx.ftz.f32 "],
            ),
        )

    def test_powf(self):
        self._test_fast_math_binary(
            pow,
            FastMathCriterion(
                fast_expected=["lg2.approx.ftz.f32 "],
                prec_unexpected=["lg2.approx.ftz.f32 "],
            ),
        )

    def test_divf(self):
        self._test_fast_math_binary(
            truediv,
            FastMathCriterion(
                fast_expected=["div.approx.ftz.f32 "],
                fast_unexpected=["div.rn.f32"],
                prec_expected=["div.rn.f32"],
                prec_unexpected=["div.approx.ftz.f32 "],
            ),
        )

    def test_divf_exception(self):
        def f10(r, x, y):
            r[0] = x / y

        sig = (float32[::1], float32, float32)
        fastver = cuda.jit(sig, fastmath=True, debug=True, opt=False)(f10)
        precver = cuda.jit(sig, debug=True, opt=False)(f10)
        nelem = 10
        ary = np.empty(nelem, dtype=np.float32)
        with self.assertRaises(ZeroDivisionError):
            precver[1, nelem](ary, 10.0, 0.0)

        try:
            fastver[1, nelem](ary, 10.0, 0.0)
        except ZeroDivisionError:
            self.fail("Divide in fastmath should not throw ZeroDivisionError")

    @unittest.expectedFailure
    def test_device_fastmath_propagation(self):
        # The fastmath option doesn't presently propagate to device functions
        # from their callees - arguably it should do, so this test is presently
        # an xfail.
        @cuda.jit("float32(float32, float32)", device=True)
        def foo(a, b):
            return a / b

        def bar(arr, val):
            i = cuda.grid(1)
            if i < arr.size:
                arr[i] = foo(i, val)

        sig = (float32[::1], float32)
        fastver = cuda.jit(sig, fastmath=True)(bar)
        precver = cuda.jit(sig)(bar)

        # Variants of the div instruction are further documented at:
        # https://docs.nvidia.com/cuda/parallel-thread-execution/index.html#floating-point-instructions-div

        # The fast version should use the "fast, approximate divide" variant
        self.assertIn("div.approx.f32", fastver.inspect_asm(sig))
        # The precise version should use the "IEEE 754 compliant rounding"
        # variant, and neither of the "approximate divide" variants.
        self.assertIn("div.rn.f32", precver.inspect_asm(sig))
        self.assertNotIn("div.approx.f32", precver.inspect_asm(sig))
        self.assertNotIn("div.full.f32", precver.inspect_asm(sig))


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import os

from llvmlite import ir
from numba.core import types, config
from numba.cuda import cgutils
from numba.core.datamodel.models import ComplexModel, UnionModel, UniTupleModel
from numba.core.debuginfo import AbstractDIBuilder
from numba.cuda.types import GridGroup

_BYTE_SIZE = 8


class DIBuilder(AbstractDIBuilder):
    DWARF_VERSION = 4
    DEBUG_INFO_VERSION = 3
    DBG_CU_NAME = "llvm.dbg.cu"
    _DEBUG = False

    def __init__(self, module, filepath, cgctx, directives_only):
        self.module = module
        self.filepath = os.path.abspath(filepath)
        self.difile = self._di_file()
        self.subprograms = []
        self.cgctx = cgctx

        if directives_only:
            self.emission_kind = "DebugDirectivesOnly"
        else:
            self.emission_kind = "FullDebug"

        self.initialize()

    def initialize(self):
        # Create the compile unit now because it is referenced when
        # constructing subprograms
        self.dicompileunit = self._di_compile_unit()

    def _var_type(self, lltype, size, datamodel=None):
        if self._DEBUG:
            print(
                "-->",
                lltype,
                size,
                datamodel,
                getattr(datamodel, "fe_type", "NO FE TYPE"),
            )
        m = self.module
        bitsize = _BYTE_SIZE * size

        int_type = (ir.IntType,)
        real_type = ir.FloatType, ir.DoubleType
        # For simple numeric types, choose the closest encoding.
        # We treat all integers as unsigned when there's no known datamodel.
        if isinstance(lltype, int_type + real_type):
            if datamodel is None:
                # This is probably something like an `i8*` member of a struct
                name = str(lltype)
                if isinstance(lltype, int_type):
                    ditok = "DW_ATE_unsigned"
                else:
                    ditok = "DW_ATE_float"
            else:
                # This is probably a known int/float scalar type
                name = str(datamodel.fe_type)
                if isinstance(datamodel.fe_type, types.Integer):
                    if datamodel.fe_type.signed:
                        ditok = "DW_ATE_signed"
                    else:
                        ditok = "DW_ATE_unsigned"
                else:
                    ditok = "DW_ATE_float"
            mdtype = m.add_debug_info(
                "DIBasicType",
                {
                    "name": name,
                    "size": bitsize,
                    "encoding": ir.DIToken(ditok),
                },
            )
        elif isinstance(datamodel, ComplexModel):
            # TODO: Is there a better way of determining "this is a complex
            # number"?
            #
            # NOTE: Commented below is the way to generate the metadata for a
            # C99 complex type that's directly supported by DWARF. Numba however
            # generates a struct with real/imag cf. CPython to give a more
            # pythonic feel to inspection.
            #
            # mdtype = m.add_debug_info('DIBasicType', {
            #  'name': f"{datamodel.fe_type} ({str(lltype)})",
            #  'size': bitsize,
            # 'encoding': ir.DIToken('DW_ATE_complex_float'),
            # })
            meta = []
            offset = 0
            for ix, name in enumerate(("real", "imag")):
                component = lltype.elements[ix]
                component_size = self.cgctx.get_abi_sizeof(component)
                component_basetype = m.add_debug_info(
                    "DIBasicType",
                    {
                        "name": str(component),
                        "size": _BYTE_SIZE * component_size,  # bits
                        "encoding": ir.DIToken("DW_ATE_float"),
                    },
                )
                derived_type = m.add_debug_info(
                    "DIDerivedType",
                    {
                        "tag": ir.DIToken("DW_TAG_member"),
                        "name": name,
                        "baseType": component_basetype,
                        "size": _BYTE_SIZE
                        * component_size,  # DW_TAG_member size is in bits
                        "offset": offset,
                    },
                )
                meta.append(derived_type)
                offset += _BYTE_SIZE * component_size  # offset is in bits
            mdtype = m.add_debug_info(
                "DICompositeType",
                {
                    "tag": ir.DIToken("DW_TAG_structure_type"),
                    "name": f"{datamodel.fe_type} ({str(lltype)})",
                    "identifier": str(lltype),
                    "elements": m.add_metadata(meta),
                    "size": offset,
                },
                is_distinct=True,
            )
        elif isinstance(datamodel, UniTupleModel):
            element = lltype.element
            el_size = self.cgctx.get_abi_sizeof(element)
            basetype = self._var_type(element, el_size)
            name = f"{datamodel.fe_type} ({str(lltype)})"
            count = size // el_size
            mdrange = m.add_debug_info(
                "DISubrange",
                {
                    "count": count,
                },
            )
            mdtype = m.add_debug_info(
                "DICompositeType",
                {
                    "tag": ir.DIToken("DW_TAG_array_type"),
                    "baseType": basetype,
                    "name": name,
                    "size": bitsize,
                    "identifier": str(lltype),
                    "elements": m.add_metadata([mdrange]),
                },
            )
        elif isinstance(lltype, ir.PointerType):
            model = getattr(datamodel, "_pointee_model", None)
            basetype = self._var_type(
                lltype.pointee, self.cgctx.get_abi_sizeof(lltype.pointee), model
            )
            mdtype = m.add_debug_info(
                "DIDerivedType",
                {
                    "tag": ir.DIToken("DW_TAG_pointer_type"),
                    "baseType": basetype,
                    "size": _BYTE_SIZE * self.cgctx.get_abi_sizeof(lltype),
                },
            )
        elif isinstance(lltype, ir.LiteralStructType):
            # Struct type
            meta = []
            offset = 0
            if datamodel is None or not datamodel.inner_models():
                name = f"Anonymous struct ({str(lltype)})"
                for field_id, element in enumerate(lltype.elements):
                    size = self.cgctx.get_abi_sizeof(element)
                    basetype = self._var_type(element, size)
                    derived_type = m.add_debug_info(
                        "DIDerivedType",
                        {
                            "tag": ir.DIToken("DW_TAG_member"),
                            "name": f"<field {field_id}>",
                            "baseType": basetype,
                            "size": _BYTE_SIZE
                            * size,  # DW_TAG_member size is in bits
                            "offset": offset,
                        },
                    )
                    meta.append(derived_type)
                    offset += _BYTE_SIZE * size  # offset is in bits
            else:
                name = f"{datamodel.fe_type} ({str(lltype)})"
                for element, field, model in zip(
                    lltype.elements, datamodel._fields, datamodel.inner_models()
                ):
                    size = self.cgctx.get_abi_sizeof(element)
                    basetype = self._var_type(element, size, datamodel=model)
                    derived_type = m.add_debug_info(
                        "DIDerivedType",
                        {
                            "tag": ir.DIToken("DW_TAG_member"),
                            "name": field,
                            "baseType": basetype,
                            "size": _BYTE_SIZE
                            * size,  # DW_TAG_member size is in bits
                            "offset": offset,
                        },
                    )
                    meta.append(derived_type)
                    offset += _BYTE_SIZE * size  # offset is in bits

            mdtype = m.add_debug_info(
                "DICompositeType",
                {
                    "tag": ir.DIToken("DW_TAG_structure_type"),
                    "name": name,
                    "identifier": str(lltype),
                    "elements": m.add_metadata(meta),
                    "size": offset,
                },
                is_distinct=True,
            )
        elif isinstance(lltype, ir.ArrayType):
            element = lltype.element
            el_size = self.cgctx.get_abi_sizeof(element)
            basetype = self._var_type(element, el_size)
            count = size // el_size
            mdrange = m.add_debug_info(
                "DISubrange",
                {
                    "count": count,
                },
            )
            mdtype = m.add_debug_info(
                "DICompositeType",
                {
                    "tag": ir.DIToken("DW_TAG_array_type"),
                    "baseType": basetype,
                    "name": str(lltype),
                    "size": bitsize,
                    "identifier": str(lltype),
                    "elements": m.add_metadata([mdrange]),
                },
            )
        else:
            # For all other types, describe it as sequence of bytes
            count = size
            mdrange = m.add_debug_info(
                "DISubrange",
                {
                    "count": count,
                },
            )
            mdbase = m.add_debug_info(
                "DIBasicType",
                {
                    "name": "byte",
                    "size": _BYTE_SIZE,
                    "encoding": ir.DIToken("DW_ATE_unsigned_char"),
                },
            )
            mdtype = m.add_debug_info(
                "DICompositeType",
                {
                    "tag": ir.DIToken("DW_TAG_array_type"),
                    "baseType": mdbase,
                    "name": str(lltype),
                    "size": bitsize,
                    "identifier": str(lltype),
                    "elements": m.add_metadata([mdrange]),
                },
            )

        return mdtype

    def mark_variable(
        self,
        builder,
        allocavalue,
        name,
        lltype,
        size,
        line,
        datamodel=None,
        argidx=None,
    ):
        arg_index = 0 if argidx is None else argidx
        m = self.module
        fnty = ir.FunctionType(ir.VoidType(), [ir.MetaDataType()] * 3)
        decl = cgutils.get_or_insert_function(m, fnty, "llvm.dbg.declare")

        mdtype = self._var_type(lltype, size, datamodel=datamodel)
        name = name.replace(".", "$")  # for gdb to work correctly
        mdlocalvar = m.add_debug_info(
            "DILocalVariable",
            {
                "name": name,
                "arg": arg_index,
                "scope": self.subprograms[-1],
                "file": self.difile,
                "line": line,
                "type": mdtype,
            },
        )
        mdexpr = m.add_debug_info("DIExpression", {})

        return builder.call(decl, [allocavalue, mdlocalvar, mdexpr])

    def mark_location(self, builder, line):
        builder.debug_metadata = self._add_location(line)

    def mark_subprogram(self, function, qualname, argnames, argtypes, line):
        name = qualname
        argmap = dict(zip(argnames, argtypes))
        di_subp = self._add_subprogram(
            name=name,
            linkagename=function.name,
            line=line,
            function=function,
            argmap=argmap,
        )
        function.set_metadata("dbg", di_subp)

    def finalize(self):
        dbgcu = cgutils.get_or_insert_named_metadata(
            self.module, self.DBG_CU_NAME
        )
        dbgcu.add(self.dicompileunit)
        self._set_module_flags()

    #
    # Internal APIs
    #

    def _set_module_flags(self):
        """Set the module flags metadata"""
        module = self.module
        mflags = cgutils.get_or_insert_named_metadata(
            module, "llvm.module.flags"
        )
        # Set *require* behavior to warning
        # See http://llvm.org/docs/LangRef.html#module-flags-metadata
        require_warning_behavior = self._const_int(2)
        if self.DWARF_VERSION is not None:
            dwarf_version = module.add_metadata(
                [
                    require_warning_behavior,
                    "Dwarf Version",
                    self._const_int(self.DWARF_VERSION),
                ]
            )
            if dwarf_version not in mflags.operands:
                mflags.add(dwarf_version)
        debuginfo_version = module.add_metadata(
            [
                require_warning_behavior,
                "Debug Info Version",
                self._const_int(self.DEBUG_INFO_VERSION),
            ]
        )
        if debuginfo_version not in mflags.operands:
            mflags.add(debuginfo_version)

    def _add_subprogram(self, name, linkagename, line, function, argmap):
        """Emit subprogram metadata"""
        subp = self._di_subprogram(name, linkagename, line, function, argmap)
        self.subprograms.append(subp)
        return subp

    def _add_location(self, line):
        """Emit location metatdaa"""
        loc = self._di_location(line)
        return loc

    @classmethod
    def _const_int(cls, num, bits=32):
        """Util to create constant int in metadata"""
        return ir.IntType(bits)(num)

    @classmethod
    def _const_bool(cls, boolean):
        """Util to create constant boolean in metadata"""
        return ir.IntType(1)(boolean)

    #
    # Helpers to emit the metadata nodes
    #

    def _di_file(self):
        return self.module.add_debug_info(
            "DIFile",
            {
                "directory": os.path.dirname(self.filepath),
                "filename": os.path.basename(self.filepath),
            },
        )

    def _di_compile_unit(self):
        return self.module.add_debug_info(
            "DICompileUnit",
            {
                "language": ir.DIToken("DW_LANG_C_plus_plus"),
                "file": self.difile,
                # Numba has to pretend to be clang to ensure the prologue is skipped
                # correctly in gdb. See:
                # https://sourceware.org/git/?p=binutils-gdb.git;a=blob;f=gdb/amd64-tdep.c;h=e563d369d8cb3eb3c2f732c2fa850ec70ba8d63b;hb=a4b0231e179607e47b1cdf1fe15c5dc25e482fad#l2521
                # Note the "producer_is_llvm" call to specialise the prologue
                # handling, this is defined here:
                # https://sourceware.org/git/?p=binutils-gdb.git;a=blob;f=gdb/producer.c;h=cdfd80d904c09394febd18749bb90359b2d128cc;hb=a4b0231e179607e47b1cdf1fe15c5dc25e482fad#l124
                # and to get a match for this condition the 'producer' must start
                # with "clang ", hence the following...
                "producer": "clang (Numba)",
                "runtimeVersion": 0,
                "isOptimized": config.OPT != 0,
                "emissionKind": ir.DIToken(self.emission_kind),
            },
            is_distinct=True,
        )

    def _di_subroutine_type(self, line, function, argmap):
        # The function call conv needs encoding.
        llfunc = function
        md = []

        for idx, llarg in enumerate(llfunc.args):
            if not llarg.name.startswith("arg."):
                name = llarg.name.replace(".", "$")  # for gdb to work correctly
                lltype = llarg.type
                size = self.cgctx.get_abi_sizeof(lltype)
                mdtype = self._var_type(lltype, size, datamodel=None)
                md.append(mdtype)

        for idx, (name, nbtype) in enumerate(argmap.items()):
            name = name.replace(".", "$")  # for gdb to work correctly
            datamodel = self.cgctx.data_model_manager[nbtype]
            lltype = self.cgctx.get_value_type(nbtype)
            size = self.cgctx.get_abi_sizeof(lltype)
            mdtype = self._var_type(lltype, size, datamodel=datamodel)
            md.append(mdtype)

        return self.module.add_debug_info(
            "DISubroutineType",
            {
                "types": self.module.add_metadata(md),
            },
        )

    def _di_subprogram(self, name, linkagename, line, function, argmap):
        return self.module.add_debug_info(
            "DISubprogram",
            {
                "name": name,
                "linkageName": linkagename,
                "scope": self.difile,
                "file": self.difile,
                "line": line,
                "type": self._di_subroutine_type(line, function, argmap),
                "isLocal": False,
                "isDefinition": True,
                "scopeLine": line,
                "isOptimized": config.OPT != 0,
                "unit": self.dicompileunit,
            },
            is_distinct=True,
        )

    def _di_location(self, line):
        return self.module.add_debug_info(
            "DILocation",
            {
                "line": line,
                "column": 1,
                "scope": self.subprograms[-1],
            },
        )


class CUDADIBuilder(DIBuilder):
    def __init__(self, module, filepath, cgctx, directives_only):
        super().__init__(module, filepath, cgctx, directives_only)
        # Cache for local variable metadata type and line deduplication
        self._vartypelinemap = {}

    def _var_type(self, lltype, size, datamodel=None):
        is_bool = False
        is_int_literal = False
        is_grid_group = False
        m = self.module

        if isinstance(lltype, ir.IntType):
            if datamodel is None:
                if size == 1:
                    name = str(lltype)
                    is_bool = True
            else:
                name = str(datamodel.fe_type)
                if isinstance(datamodel.fe_type, types.Boolean):
                    is_bool = True
                    if isinstance(datamodel.fe_type, types.BooleanLiteral):
                        name = "bool"
                elif isinstance(datamodel.fe_type, types.Integer):
                    if isinstance(datamodel.fe_type, types.IntegerLiteral):
                        name = f"int{_BYTE_SIZE * size}"
                        is_int_literal = True
                elif isinstance(datamodel.fe_type, GridGroup):
                    is_grid_group = True

        if is_bool or is_int_literal or is_grid_group:
            bitsize = _BYTE_SIZE * size
            # Boolean type workaround until upstream Numba is fixed
            if is_bool:
                ditok = "DW_ATE_boolean"
            elif is_int_literal:
                ditok = "DW_ATE_signed"
            # GridGroup type should use numba.cuda implementation
            elif is_grid_group:
                ditok = "DW_ATE_unsigned"

            return m.add_debug_info(
                "DIBasicType",
                {
                    "name": name,
                    "size": bitsize,
                    "encoding": ir.DIToken(ditok),
                },
            )

        if isinstance(datamodel, UnionModel):
            # UnionModel is handled here to represent polymorphic types
            meta = []
            maxwidth = 0
            for field, model in zip(
                datamodel._fields, datamodel.inner_models()
            ):
                # Ignore the "tag" field, focus on the "payload" field which
                # contains the data types in memory
                if field == "payload":
                    for mod in model.inner_models():
                        dtype = mod.get_value_type()
                        membersize = self.cgctx.get_abi_sizeof(dtype)
                        basetype = self._var_type(
                            dtype, membersize, datamodel=mod
                        )
                        if isinstance(mod.fe_type, types.Literal):
                            typename = str(mod.fe_type.literal_type)
                        else:
                            typename = str(mod.fe_type)
                        # Use a prefix "_" on type names as field names
                        membername = "_" + typename
                        memberwidth = _BYTE_SIZE * membersize
                        derived_type = m.add_debug_info(
                            "DIDerivedType",
                            {
                                "tag": ir.DIToken("DW_TAG_member"),
                                "name": membername,
                                "baseType": basetype,
                                # DW_TAG_member size is in bits
                                "size": memberwidth,
                            },
                        )
                        meta.append(derived_type)
                        if memberwidth > maxwidth:
                            maxwidth = memberwidth

            fake_union_name = "dbg_poly_union"
            return m.add_debug_info(
                "DICompositeType",
                {
                    "file": self.difile,
                    "tag": ir.DIToken("DW_TAG_union_type"),
                    "name": fake_union_name,
                    "identifier": str(lltype),
                    "elements": m.add_metadata(meta),
                    "size": maxwidth,
                },
                is_distinct=True,
            )
        # For other cases, use upstream Numba implementation
        return super()._var_type(lltype, size, datamodel=datamodel)

    def _di_subroutine_type(self, line, function, argmap):
        # The function call conv needs encoding.
        llfunc = function
        md = []

        # Create metadata type for return value
        if len(llfunc.args) > 0:
            lltype = llfunc.args[0].type
            size = self.cgctx.get_abi_sizeof(lltype)
            mdtype = self._var_type(lltype, size, datamodel=None)
            md.append(mdtype)

        # Create metadata type for arguments
        for idx, (name, nbtype) in enumerate(argmap.items()):
            datamodel = self.cgctx.data_model_manager[nbtype]
            lltype = self.cgctx.get_value_type(nbtype)
            size = self.cgctx.get_abi_sizeof(lltype)
            mdtype = self._var_type(lltype, size, datamodel=datamodel)
            md.append(mdtype)

        return self.module.add_debug_info(
            "DISubroutineType",
            {
                "types": self.module.add_metadata(md),
            },
        )

    def mark_variable(
        self,
        builder,
        allocavalue,
        name,
        lltype,
        size,
        line,
        datamodel=None,
        argidx=None,
    ):
        if name.startswith("$") or "." in name:
            # Do not emit llvm.dbg.declare on user variable alias
            return
        else:
            int_type = (ir.IntType,)
            real_type = ir.FloatType, ir.DoubleType
            if isinstance(lltype, int_type + real_type):
                # Start with scalar variable, swtiching llvm.dbg.declare
                # to llvm.dbg.value
                return
            else:
                return super().mark_variable(
                    builder,
                    allocavalue,
                    name,
                    lltype,
                    size,
                    line,
                    datamodel,
                    argidx,
                )

    def update_variable(
        self,
        builder,
        value,
        name,
        lltype,
        size,
        line,
        datamodel=None,
        argidx=None,
    ):
        m = self.module
        fnty = ir.FunctionType(ir.VoidType(), [ir.MetaDataType()] * 3)
        decl = cgutils.get_or_insert_function(m, fnty, "llvm.dbg.value")

        mdtype = self._var_type(lltype, size, datamodel)
        index = name.find(".")
        if index >= 0:
            name = name[:index]
        # Merge DILocalVariable nodes with same name and type but different
        # lines. Use the cached [(name, type) -> line] info to deduplicate
        # metadata. Use the lltype as part of key.
        key = (name, lltype)
        if key in self._vartypelinemap:
            line = self._vartypelinemap[key]
        else:
            self._vartypelinemap[key] = line
        arg_index = 0 if argidx is None else argidx
        mdlocalvar = m.add_debug_info(
            "DILocalVariable",
            {
                "name": name,
                "arg": arg_index,
                "scope": self.subprograms[-1],
                "file": self.difile,
                "line": line,
                "type": mdtype,
            },
        )
        mdexpr = m.add_debug_info("DIExpression", {})

        return builder.call(decl, [value, mdlocalvar, mdexpr])
